from pathlib import Path

import numpy
import pytest
from openff.toolkit import Molecule, Quantity, Topology
from openff.utilities.testing import skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import HAS_GROMACS, get_test_files_dir_path, needs_gmx
from openff.interchange.components.potentials import Potential
from openff.interchange.constants import kj_mol
from openff.interchange.drivers import get_openmm_energies
from openff.interchange.foyer._guard import has_foyer
from openff.interchange.models import PotentialKey, TopologyKey

if has_foyer:
    import foyer

    from openff.interchange.foyer._valence import _RBTorsionHandler

if HAS_GROMACS:
    from openff.interchange.drivers.gromacs import (
        _get_mdp_file,
        _process,
        _run_gmx_energy,
        get_gromacs_energies,
    )


@pytest.mark.skipif(not has_foyer, reason="Foyer is not installed")
class TestFoyer:
    @pytest.fixture
    def oplsaa(self):
        return foyer.forcefields.load_OPLSAA()

    @pytest.fixture
    def oplsaa_interchange_ethanol(self, oplsaa):
        molecule = Molecule.from_file(
            get_test_files_dir_path("foyer_test_molecules") / "ethanol.sdf",
        )
        molecule.name = "ETH"

        top = molecule.to_topology()

        interchange = Interchange.from_foyer(topology=top, force_field=oplsaa)
        interchange.positions = molecule.conformers[0].m_as("nanometer")
        interchange.box = [4, 4, 4]
        return interchange

    @pytest.fixture
    def get_interchanges(self, oplsaa):
        def interchanges_from_path(molecule_path):
            pmd = pytest.importorskip("parmed")

            molecule_or_molecules = Molecule.from_file(molecule_path)
            mol_name = Path(molecule_path).stem[0:4].upper()

            if isinstance(molecule_or_molecules, list):
                for idx, mol in enumerate(molecule_or_molecules):
                    mol.name = f"{mol_name}{idx}"
            else:
                molecule_or_molecules.name = mol_name
                molecule_or_molecules = [molecule_or_molecules]

            topology = Topology.from_molecules(molecule_or_molecules)
            openff_interchange = Interchange.from_foyer(oplsaa, topology)

            if isinstance(molecule_or_molecules, list):
                openff_interchange.positions = numpy.vstack(
                    tuple(molecule.conformers[0].m_as("nanometer") for molecule in molecule_or_molecules),
                )
            else:
                openff_interchange.positions = molecule_or_molecules.conformers[0].m_as(
                    "nanometer",
                )

            openff_interchange.box = [4, 4, 4]

            parmed_struct = pmd.openmm.load_topology(topology.to_openmm())
            parmed_struct.positions = openff_interchange.positions.m_as("angstrom")
            parmed_struct.box = [40, 40, 40, 90, 90, 90]

            return openff_interchange, parmed_struct

        return interchanges_from_path

    def test_handlers_exist(self, oplsaa_interchange_ethanol):
        for _, handler in oplsaa_interchange_ethanol.collections.items():
            assert handler

        assert oplsaa_interchange_ethanol["vdW"].scale_14 == 0.5
        assert oplsaa_interchange_ethanol["Electrostatics"].scale_14 == 0.5

    @needs_gmx
    @pytest.mark.slow
    @pytest.mark.skip(
        reason="Exporting to OpenMM with geometric mixing rules not yet implemented",
    )
    def test_ethanol_energies(self, oplsaa_interchange_ethanol):
        from openff.interchange.drivers import get_gromacs_energies

        gmx_energies = get_gromacs_energies(oplsaa_interchange_ethanol)
        omm_energies = get_openmm_energies(oplsaa_interchange_ethanol)

        gmx_energies.compare(
            omm_energies,
            {
                "vdW": Quantity(12.0, "kilojoule / mole"),
                "Electrostatics": Quantity(12.0, "kilojoule / mole"),
            },
        )

    @needs_gmx
    @pytest.mark.filterwarnings("ignore:Parameters have not been assigned to all impropers.")
    @pytest.mark.parametrize(
        argnames="molecule_path",
        argvalues=[*get_test_files_dir_path("foyer_test_molecules").glob("*.sdf")],
    )
    def test_interchange_energies(self, molecule_path, get_interchanges, oplsaa):
        if "ethanol" in str(molecule_path) or "adamantane" in str(molecule_path):
            pytest.skip("Foyer/ParmEd bug with this molecule")
        openff_interchange, pmd_structure = get_interchanges(molecule_path)
        parameterized_pmd_structure = oplsaa.apply(pmd_structure)

        openff_energy = get_gromacs_energies(
            openff_interchange,
            mdp="cutoff_hbonds",
            round_positions=3,
        )

        parameterized_pmd_structure.save("from_foyer.gro")
        parameterized_pmd_structure.save("from_foyer.top")

        through_foyer = _process(
            _run_gmx_energy(
                top_file="from_foyer.top",
                gro_file="from_foyer.gro",
                mdp_file=_get_mdp_file("cutoff_hbonds"),
            ),
        )

        # TODO: Revisit after https://github.com/mosdef-hub/foyer/issues/431
        openff_energy.compare(
            through_foyer,
            {
                "Bond": Quantity(1e-3, "kilojoule / mole"),
                "Angle": Quantity(1e-3, "kilojoule / mole"),
                "Torsion": Quantity(1e-3, "kilojoule / mole"),
                "vdW": Quantity(5, "kilojoule / mole"),
                "Electrostatics": Quantity(1, "kilojoule / mole"),
            },
        )


class TestRBTorsions(TestFoyer):
    @pytest.fixture
    def ethanol_with_rb_torsions(self, sage):
        mol = Molecule.from_smiles("CC")
        mol.name = "ETH"
        mol.generate_conformers(n_conformers=1)
        top = mol.to_topology()

        out = Interchange.from_smirnoff(sage, top)
        out.box = [4, 4, 4]
        out.positions = mol.conformers[0]
        out.positions = numpy.round(out.positions, 2)

        rb_torsions = _RBTorsionHandler()
        smirks = "[#1:1]-[#6X4:2]-[#6X4:3]-[#1:4]"
        pot_key = PotentialKey(id=smirks)
        for proper in top.propers:
            top_key = TopologyKey(atom_indices=tuple(top.atom_index(a) for a in proper))
            rb_torsions.key_map.update({top_key: pot_key})

        # Values from HC-CT-CT-HC RB torsion
        # https://github.com/mosdef-hub/foyer/blob/7816bf53a127502520a18d76c81510f96adfdbed/foyer/forcefields/xml/oplsaa.xml#L2585
        pot = Potential(
            parameters={
                "c0": 0.6276 * kj_mol,
                "c1": 1.8828 * kj_mol,
                "c2": 0.0 * kj_mol,
                "c3": -2.5104 * kj_mol,
                "c4": 0.0 * kj_mol,
                "c5": 0.0 * kj_mol,
            },
        )

        rb_torsions.potentials.update({pot_key: pot})

        out.collections.update({"RBTorsions": rb_torsions})
        out.collections.pop("ProperTorsions")

        return out

    @needs_gmx
    @pytest.mark.slow
    @pytest.mark.skip(reason="Something is broken with RBTorsions in OpenMM export")
    def test_rb_torsions(self, ethanol_with_rb_torsions):
        omm = get_openmm_energies(ethanol_with_rb_torsions, round_positions=3).energies["Torsion"]
        gmx = get_gromacs_energies(
            ethanol_with_rb_torsions,
            round_positions=3,
        ).energies["Torsion"]

        assert (gmx - omm).m_as(kj_mol) < 1e-6

    @pytest.mark.filterwarnings("ignore:Parameters have not been assigned to all impropers.")
    @skip_if_missing("mbuild")
    @needs_gmx
    def test_rb_torsions_vs_foyer(self, oplsaa, ethanol_with_rb_torsions):
        # Given that these force constants are copied from Foyer's OPLS-AA file,
        # compare to processing through the current MoSDeF pipeline
        import mbuild

        comp = mbuild.load("CC", smiles=True)
        comp.xyz = ethanol_with_rb_torsions.positions.m_as("nanometer")

        from_foyer = oplsaa.apply(comp)
        from_foyer.box = [40, 40, 40, 90, 90, 90]
        from_foyer.save("from_foyer.top")
        from_foyer.save("from_foyer.gro")

        rb_torsion_energy_from_foyer = _process(
            _run_gmx_energy(
                top_file="from_foyer.top",
                gro_file="from_foyer.gro",
                mdp_file=_get_mdp_file("default"),
            ),
        ).energies["Torsion"]

        # GROMACS vs. OpenMM was already compared, so just use one
        omm = get_gromacs_energies(
            ethanol_with_rb_torsions,
            round_positions=3,
        ).energies["Torsion"]

        assert (omm - rb_torsion_energy_from_foyer).m_as(kj_mol) < 1e-6



import pytest
from openff.toolkit import Molecule, Quantity
from openff.utilities.testing import skip_if_missing

from openff.interchange.exceptions import (
    UnsupportedCutoffMethodError,
    UnsupportedExportError,
)


@skip_if_missing("openmm")
class TestUnsupportedCases:
    def test_ljpme_nonperiodic(self, sage):
        interchange = sage.create_interchange(Molecule.from_smiles("CC").to_topology())

        interchange["vdW"].nonperiodic_method = "pme"

        with pytest.raises(
            UnsupportedCutoffMethodError,
            match="not valid for non-periodic systems",
        ):
            interchange.to_openmm(combine_nonbonded_forces=False)

    @pytest.mark.parametrize("periodic", [True, False])
    @pytest.mark.parametrize("combine", [True, False])
    def test_hard_cutoff(self, sage, periodic, combine):
        interchange = sage.create_interchange(Molecule.from_smiles("CC").to_topology())

        if periodic:
            interchange.box = Quantity([4, 4, 4], "nanometer")
            interchange["Electrostatics"].periodic_potential = "cutoff"
        else:
            interchange["Electrostatics"].nonperiodic_potential = "cutoff"

        with pytest.raises(
            UnsupportedCutoffMethodError,
            match=r"does not support.*Consider using",
        ):
            interchange.to_openmm(combine_nonbonded_forces=combine)


@skip_if_missing("openmm")
class TestCutoffElectrostatics:
    @pytest.mark.parametrize("periodic", [True, False])
    @pytest.mark.parametrize("combine", [True, False])
    def test_export_reaction_field_electrostatics(
        self,
        sage,
        basic_top,
        periodic,
        combine,
    ):
        import openmm
        import openmm.unit

        out = sage.create_interchange(basic_top)

        if periodic:
            out["Electrostatics"].periodic_potential = "reaction-field"
        else:
            out["Electrostatics"].nonperiodic_potential = "reaction-field"
            out.box = None

        assert (out.box is not None) == periodic

        out["Electrostatics"].cutoff = out["vdW"].cutoff

        if combine and not periodic:
            with pytest.raises(
                UnsupportedCutoffMethodError,
                match="Combination of",
            ):
                system = out.to_openmm(combine_nonbonded_forces=combine)
            return

        system = out.to_openmm(combine_nonbonded_forces=combine)

        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                assert 0.9 == force.getCutoffDistance() / openmm.unit.nanometer

                if periodic:
                    expected_force = openmm.NonbondedForce.CutoffPeriodic
                else:
                    expected_force = openmm.NonbondedForce.CutoffNonPeriodic

                assert expected_force == force.getNonbondedMethod()

                del force

                break
        else:
            pytest.fail("Found no `NonbondedForce`")

    def test_vdw_electrostatics_cutoff_mismatch(
        self,
        sage,
        basic_top,
    ):
        import random

        out = sage.create_interchange(basic_top)

        out["Electrostatics"].periodic_potential = "reaction-field"
        out["Electrostatics"].cutoff = out["vdW"].cutoff

        for key in ("vdW", "Electrostatics"):
            out[key].cutoff *= random.random()

        with pytest.raises(
            UnsupportedExportError,
            match="cutoffs must match",
        ):
            out.to_openmm(combine_nonbonded_forces=True)


@skip_if_missing("openmm")
class TestEwaldSettings:
    @pytest.mark.parametrize("lj_method", ["cutoff", "pme"])
    def test_set_ewald_tolerance(self, sage, basic_top, lj_method):
        import openmm

        if lj_method == "pme":
            sage["vdW"].periodic_method = "Ewald3D"

        system = sage.create_interchange(basic_top).to_openmm(ewald_tolerance=1.234e-5)

        for force in system.getForces():
            if isinstance(force, openmm.NonbondedForce):
                if lj_method == "pme":
                    assert force.getNonbondedMethod() == openmm.NonbondedForce.LJPME
                elif lj_method == "cutoff":
                    assert force.getNonbondedMethod() == openmm.NonbondedForce.PME

                assert force.getEwaldErrorTolerance() == 1.234e-5

                break
        else:
            pytest.fail("Found no `NonbondedForce`")



from openff.interchange._tests import get_test_file_path
from openff.interchange.interop.gromacs._import._import import GROMACSSystem
from openff.interchange.interop.gromacs._import._topology import (
    _create_topology_from_system,
)


def test_complex(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    topology = _create_topology_from_system(
        GROMACSSystem.from_files(
            get_test_file_path("complex.top"),
            get_test_file_path("complex.gro"),
        ),
    )

    assert topology.n_molecules == 2165
    assert topology.n_atoms == 6656

    assert topology.molecule(0).n_atoms == 144
    assert topology.molecule(1).n_atoms == 49
    assert topology.molecule(2).n_atoms == 1
    assert topology.molecule(15).n_atoms == 3
    assert topology.molecule(-1).n_atoms == 3

    assert topology.molecule(2).n_bonds == 0
    assert topology.molecule(15).n_bonds == 2
    assert topology.molecule(-1).n_bonds == 2



from openff.utilities import has_package, skip_if_missing

if has_package("openmm"):
    import numpy
    import pytest
    from openff.toolkit import Molecule

    from openff.interchange._tests import MoleculeWithConformer
    from openff.interchange.drivers import get_openmm_energies
    from openff.interchange.exceptions import MissingPositionsError
    from openff.interchange.operations.minimize import (
        _DEFAULT_ENERGY_MINIMIZATION_TOLERANCE,
    )
    from openff.interchange.operations.minimize.openmm import minimize_openmm


@skip_if_missing("openmm")
class TestOpenMMMinimization:
    @pytest.mark.slow
    def test_minimization_decreases_energy(self, sage):
        system = sage.create_interchange(
            MoleculeWithConformer.from_smiles("CNC(=O)C1=C(N)C=C(F)C=C1").to_topology(),
        )

        original_energy = get_openmm_energies(system).total_energy

        minimized_positions = minimize_openmm(
            system,
            tolerance=_DEFAULT_ENERGY_MINIMIZATION_TOLERANCE,
            max_iterations=10_000,
        )

        assert not numpy.allclose(minimized_positions, system.positions)

        system.positions = minimized_positions
        minimied_energy = get_openmm_energies(system).total_energy

        assert minimied_energy < original_energy

    def test_missing_positions_error(self, tip3p):
        with pytest.raises(MissingPositionsError, match="positions=None"):
            tip3p.create_interchange(Molecule.from_smiles("O").to_topology()).minimize()

    def test_minimization_does_not_add_virtual_sites_as_atoms(self, tip4p, water_tip4p):
        system = tip4p.create_interchange(water_tip4p.to_topology())

        original_positions = system.positions

        system.minimize()

        assert system.positions.shape == original_positions.shape



"""Fixtures for driver tests."""

from copy import deepcopy

import pytest
from openff.toolkit import Quantity

from openff.interchange._tests import MoleculeWithConformer


@pytest.fixture
def methane_dimer(sage):
    molecule = MoleculeWithConformer.from_smiles("C")
    topology = deepcopy(molecule).to_topology()

    molecule._conformers[0] += Quantity([4, 0, 0], "angstrom")

    topology.add_molecule(molecule)
    topology.box_vectors = Quantity([3, 3, 3], "nanometer")

    interchange = sage.create_interchange(topology)
    interchange.minimize()

    return interchange



"""
Helper functions for producing `openmm.Force` objects for non-bonded terms.
"""

import itertools
import warnings
from collections import defaultdict
from typing import NamedTuple

from openff.toolkit import Molecule, Quantity
from openff.units.openmm import to_openmm as to_openmm_quantity
from openff.utilities.utilities import has_package

from openff.interchange import Interchange
from openff.interchange.common._nonbonded import ElectrostaticsCollection, _simpler_charges, vdWCollection
from openff.interchange.constants import _PME
from openff.interchange.exceptions import (
    CannotSetSwitchingFunctionError,
    InternalInconsistencyError,
    UnsupportedCutoffMethodError,
    UnsupportedExportError,
)
from openff.interchange.interop.common import _build_particle_map
from openff.interchange.models import (
    BaseVirtualSiteKey,
    ImportedVirtualSiteKey,
    SingleAtomChargeTopologyKey,
    TopologyKey,
)
from openff.interchange.warnings import NonbondedSettingsWarning

if has_package("openmm"):
    import openmm
    import openmm.unit

# TODO: Currently, these are not used since the openmm.CustomNonbondedForce does not handle 1-4 interactions and
#       instead they are handleded by an openmm.CustomBondForce in which the scaled parameters are manually computed.
_MIXING_RULE_EXPRESSIONS: dict[str, str] = {
    "lorentz-berthelot": "sigma=(sigma1+sigma2)/2; epsilon=sqrt(epsilon1*epsilon2); ",
    "geometric": "sigma=sqrt(sigma1*sigma2); epsilon=sqrt(epsilon1*epsilon2); ",
}


class _NonbondedData(NamedTuple):
    vdw_collection: vdWCollection
    vdw_cutoff: Quantity
    vdw_method: str | None
    vdw_expression: str | None
    mixing_rule: str
    electrostatics_collection: ElectrostaticsCollection
    electrostatics_method: str | None
    periodic: bool


def _process_nonbonded_forces(
    interchange: "Interchange",
    system: openmm.System,
    combine_nonbonded_forces: bool = False,
    ewald_tolerance: float = 1e-4,
) -> dict[int | BaseVirtualSiteKey, int]:
    """
    Process the non-bonded collections in an Interchange into corresponding openmm objects.

    This typically involves processing the vdW and Electrostatics sections of an Interchange object
    into a corresponding openmm.NonbondedForce (if `combine_nonbonded_forces=True`) or a
    collection of other forces (NonbondedForce, CustomNonbondedForce, CustomBondForce) if
    `combine_nonbonded_forces=False`.

    """
    if "VirtualSites" in interchange.collections:
        from openff.interchange.interop._virtual_sites import (
            _virtual_site_parent_molecule_mapping,
        )
        from openff.interchange.interop.common import (
            _check_virtual_site_exclusion_policy,
        )

        virtual_sites = interchange["VirtualSites"]

        _check_virtual_site_exclusion_policy(virtual_sites)

        virtual_site_molecule_map: dict[
            BaseVirtualSiteKey,
            int,
        ] = _virtual_site_parent_molecule_mapping(interchange)

        molecule_virtual_site_map: dict[int, list[BaseVirtualSiteKey]] = defaultdict(list)

        for virtual_site_key, molecule_index in virtual_site_molecule_map.items():
            molecule_virtual_site_map[molecule_index].append(virtual_site_key)

    else:
        molecule_virtual_site_map = defaultdict(list)

    # Mapping between OpenFF "particles" and OpenMM particles (via index). OpenFF objects
    # (keys) are either atom indices (if atoms) or `VirtualSiteKey`s if virtual sites
    # openff_openmm_particle_map: dict[Union[int, VirtualSiteKey], int] = dict()
    openff_openmm_particle_map = _add_particles_to_system(
        interchange,
        system,
        molecule_virtual_site_map,
    )

    # TODO: Process ElectrostaticsHandler.exception_potential

    _data = _prepare_input_data(interchange)

    if _data.vdw_collection is None and _data.electrostatics_collection is None:
        # Will not make any non-bonded forces, but we have made particles, so just return early
        return openff_openmm_particle_map

    if combine_nonbonded_forces:
        _func = _create_single_nonbonded_force
    else:
        _func = _create_multiple_nonbonded_forces

    _func(
        _data,
        interchange,
        system,
        ewald_tolerance,
        molecule_virtual_site_map,
        openff_openmm_particle_map,
    )

    return openff_openmm_particle_map


def _add_particles_to_system(
    interchange: "Interchange",
    system: openmm.System,
    molecule_virtual_site_map,
) -> dict[int | BaseVirtualSiteKey, int]:
    particle_map = _build_particle_map(
        interchange,
        molecule_virtual_site_map,
        collate=False,
    )

    for molecule in interchange.topology.molecules:
        for atom in molecule.atoms:
            atom_index = interchange.topology.atom_index(atom)

            # Skip unit check for speed, trust the toolkit reports mass in Dalton
            system_index = system.addParticle(mass=atom.mass.m)

            assert system_index == particle_map[atom_index], (
                system_index,
                atom_index,
                particle_map[atom_index],
            )

    for molecule in interchange.topology.molecules:
        for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(molecule)]:
            from openff.interchange.interop.openmm._virtual_sites import (
                _create_openmm_virtual_site,
            )
            from openff.interchange.smirnoff._virtual_sites import (
                _create_virtual_site_object,
            )

            system_index = system.addParticle(mass=0.0)

            assert system_index == particle_map[virtual_site_key]

            virtual_site_potential = interchange["VirtualSites"].potentials[
                interchange["VirtualSites"].key_map[virtual_site_key]
            ]

            virtual_site_object = _create_virtual_site_object(
                virtual_site_key,
                virtual_site_potential,
            )

            openmm_particle: openmm.VirtualSite = _create_openmm_virtual_site(
                interchange,
                virtual_site_object,
                particle_map,
            )

            system.setVirtualSite(system_index, openmm_particle)

    return particle_map


def _prepare_input_data(interchange: "Interchange") -> _NonbondedData:
    try:
        vdw: vdWCollection = interchange["vdW"]
    except LookupError:
        for collection in interchange.collections.values():
            if collection.is_plugin:
                if collection.acts_as == "vdW":  # type: ignore[attr-defined]
                    # We can't be completely sure all plugins subclass out of vdWCollection here
                    vdw = collection  # type: ignore[assignment]
                    break
        else:
            vdw = None  # type: ignore[assignment]

    if vdw:
        vdw_cutoff: Quantity | None = vdw.cutoff

        if interchange.box is None:
            vdw_method: str | None = vdw.nonperiodic_method.lower()
        else:
            vdw_method = vdw.periodic_method.lower()

        mixing_rule: str = getattr(vdw, "mixing_rule", "")
        vdw_expression: str | None = vdw.expression.replace("**", "^")
    else:
        vdw_cutoff = None
        vdw_method = None
        mixing_rule = ""
        vdw_expression = None

    try:
        electrostatics: ElectrostaticsCollection = interchange["Electrostatics"]
    except LookupError:
        electrostatics = None  # type: ignore[assignment]

    if electrostatics is None:
        electrostatics_method: str | None = None
    else:
        if interchange.box is None:
            electrostatics_method = getattr(
                electrostatics,
                "nonperiodic_potential",
                "Coulomb",
            )
        else:
            electrostatics_method = getattr(electrostatics, "periodic_potential", _PME)

    return _NonbondedData(
        vdw_collection=vdw,
        vdw_cutoff=vdw_cutoff,
        vdw_method=vdw_method,
        vdw_expression=vdw_expression,
        mixing_rule=mixing_rule,
        electrostatics_collection=electrostatics,
        electrostatics_method=electrostatics_method,
        periodic=interchange.box is not None,
    )


def _create_single_nonbonded_force(
    data: _NonbondedData,
    interchange: "Interchange",
    system: openmm.System,
    ewald_tolerance: float,
    molecule_virtual_site_map: dict["Molecule", list[BaseVirtualSiteKey]],
    openff_openmm_particle_map: dict[int | BaseVirtualSiteKey, int],
):
    """Create a single openmm.NonbondedForce from vdW/electrostatics/virtual site collections."""
    if data.mixing_rule not in ("lorentz-berthelot", ""):
        raise UnsupportedExportError(
            "OpenMM's default NonbondedForce only supports Lorentz-Berthelot mixing rules. "
            f"Found {data.mixing_rule}. "
            "Try setting `combine_nonbonded_forces=False`.",
        )

    if molecule_virtual_site_map in (None, dict()):
        has_virtual_sites = False
    elif all([len(v) == 0 for v in molecule_virtual_site_map.values()]):
        has_virtual_sites = False
    else:
        has_virtual_sites = True

    non_bonded_force = openmm.NonbondedForce()
    non_bonded_force.setName("Nonbonded force")
    system.addForce(non_bonded_force)

    # This limitation is independent of periodicity and vdW method, so check it first
    if data.electrostatics_method == "cutoff":
        raise UnsupportedCutoffMethodError(
            "OpenMM does not support electrostatics with a hard cutoff (and no reaction field "
            'modification). Consider using `"reaction-field"` to get force(s) with CutoffPeriodic '
            "or CutoffNonperiodic.",
        )

    if interchange.box is None:
        if (data.vdw_method in ("no-cutoff", None)) and (data.electrostatics_method in ("Coulomb", None)):
            non_bonded_force.setNonbondedMethod(openmm.NonbondedForce.NoCutoff)
            non_bonded_force.setUseDispersionCorrection(True)
            if data.vdw_cutoff:
                non_bonded_force.setCutoffDistance(
                    to_openmm_quantity(data.vdw_cutoff),
                )
        else:
            raise UnsupportedCutoffMethodError(
                f"Combination of non-bonded cutoff methods {data.vdw_method} (vdW) and "
                f"{data.electrostatics_method} (Electrostatics) not currently supported or "
                f"invalid with `combine_nonbonded_forces=True` and `.box={interchange.box}`.",
            )

    else:
        if data.vdw_method in ("cutoff", None) and data.electrostatics_method in (
            _PME,
            None,
        ):
            non_bonded_force.setNonbondedMethod(openmm.NonbondedForce.PME)
            non_bonded_force.setEwaldErrorTolerance(ewald_tolerance)
            non_bonded_force.setUseDispersionCorrection(True)
            if not data.vdw_cutoff:
                # With no vdW handler and/or ambiguous cutoff, cannot set it,
                # thereforce silently fall back to OpenMM's default. It's not
                # clear if this value matters with only (PME) charges and no
                # vdW interactions in the system.
                pass
            else:
                non_bonded_force.setCutoffDistance(
                    to_openmm_quantity(data.vdw_cutoff),
                )
        elif data.vdw_method == "pme" and data.electrostatics_method == _PME:
            non_bonded_force.setNonbondedMethod(openmm.NonbondedForce.LJPME)
            non_bonded_force.setEwaldErrorTolerance(ewald_tolerance)

        elif (data.vdw_method == "cutoff") and (data.electrostatics_method == "reaction-field"):
            if data.vdw_cutoff != data.electrostatics_collection.cutoff:
                raise UnsupportedExportError(
                    "If using cutoff vdW and reaction-field electrostatics, cutoffs must match.",
                )

            non_bonded_force.setNonbondedMethod(openmm.NonbondedForce.CutoffPeriodic)
            non_bonded_force.setCutoffDistance(
                to_openmm_quantity(data.vdw_cutoff),
            )

        else:
            raise UnsupportedCutoffMethodError(
                f"Combination of non-bonded cutoff methods {data.vdw_method} (vdW) and "
                f"{data.electrostatics_method} (Electrostatics) not currently supported or invalid with "
                f"`combine_nonbonded_forces=True` and `.box={interchange.box}`.",
            )

    if data.electrostatics_collection is not None:
        partial_charges = _simpler_charges(data.electrostatics_collection)

    # mapping between (openmm) index of each atom and the (openmm) index of each virtual particle
    #   of that parent atom (if any)
    # if no virtual sites at all, this remains an empty dict
    parent_virtual_particle_mapping: defaultdict[int, list[int]] = defaultdict(list)

    vdw = data.vdw_collection

    for molecule in interchange.topology.molecules:
        for atom in molecule.atoms:
            non_bonded_force.addParticle(0.0, 1.0, 0.0)

            atom_index = interchange.topology.atom_index(atom)

            top_key = TopologyKey(atom_indices=(atom_index,))

            if data.electrostatics_collection is not None:
                try:
                    partial_charge = partial_charges[top_key].m_as("elementary_charge")
                except KeyError:
                    # TODO: Work around this by updating the handler or .charges
                    #       to support looking up directly based on atom index,
                    #       not creating a new TopologyKey each time
                    other_top_key = SingleAtomChargeTopologyKey(
                        this_atom_index=atom_index,
                    )

                    partial_charge = partial_charges[other_top_key].m_as(  # type: ignore[index]
                        "elementary_charge",
                    )

            else:
                partial_charge = 0.0

            if vdw is not None:
                pot_key = vdw.key_map[top_key]

                sigma = vdw.potentials[pot_key].parameters["sigma"]
                epsilon = vdw.potentials[pot_key].parameters["epsilon"]

                sigma = sigma.m_as("nanometer")
                epsilon = epsilon.m_as("kilojoule / mole")
            else:
                sigma = openmm.unit.Quantity(0.0, openmm.unit.nanometer)
                epsilon = openmm.unit.Quantity(0.0, openmm.unit.kilojoules_per_mole)

            openmm_atom_index = openff_openmm_particle_map[atom_index]

            non_bonded_force.setParticleParameters(
                openmm_atom_index,
                partial_charge,
                sigma,
                epsilon,
            )

    for molecule in interchange.topology.molecules:
        if has_virtual_sites:
            molecule_index = interchange.topology.molecule_index(molecule)
        else:
            continue

        for virtual_site_key in molecule_virtual_site_map[molecule_index]:
            # TODO: Move this function to openff/interchange/interop/_particles.py ?
            from openff.interchange.smirnoff._virtual_sites import (
                _create_virtual_site_object,
            )

            _potential_key = interchange["VirtualSites"].key_map[virtual_site_key]
            virtual_site_potential = interchange["VirtualSites"].potentials[_potential_key]
            virtual_site_object = _create_virtual_site_object(
                virtual_site_key,
                virtual_site_potential,
            )

            vdw = interchange["vdW"]
            coul = interchange["Electrostatics"]

            vdw_key = vdw.key_map.get(virtual_site_key)
            coul_key = coul.key_map.get(virtual_site_key)
            if vdw_key is None or coul_key is None:
                raise InternalInconsistencyError(
                    f"Virtual site {virtual_site_key} is not associated with any "
                    "vdW and/or electrostatics interactions",
                )

            try:
                # SMIRNOFF-style virtual sites are charged from charge increments
                # moved from their parent atoms
                charge_increments = coul.potentials[coul_key].parameters["charge_increments"]

                charge = to_openmm_quantity(-sum(charge_increments))
            except KeyError:
                # but stuff imported from other sources just has the charge right there
                assert isinstance(virtual_site_key, ImportedVirtualSiteKey)

                charge = coul.potentials[coul_key].parameters["charge"].to_openmm()

            vdw_parameters = vdw.potentials[vdw_key].parameters
            sigma = to_openmm_quantity(vdw_parameters["sigma"])
            epsilon = to_openmm_quantity(vdw_parameters["epsilon"])

            system_index: int = openff_openmm_particle_map[virtual_site_key]
            force_index = non_bonded_force.addParticle(charge, sigma, epsilon)

            if system_index != force_index:
                raise InternalInconsistencyError(
                    f"Mismatch in system ({system_index=}) and force ({force_index=}) indexing",
                )

            try:
                parent_atom_index = openff_openmm_particle_map[virtual_site_object.orientations[0]]
            except AttributeError:
                # TODO: Should unify these objects
                assert isinstance(virtual_site_key, ImportedVirtualSiteKey)

                parent_atom_index = openff_openmm_particle_map[virtual_site_object.particles[0]]  # type: ignore[attr-defined]

            parent_virtual_particle_mapping[parent_atom_index].append(force_index)

    _create_exceptions(
        data,
        non_bonded_force,
        interchange,
        openff_openmm_particle_map,
        parent_virtual_particle_mapping,
    )

    _apply_switching_function(data.vdw_collection, non_bonded_force)


def _create_exceptions(
    data: _NonbondedData,
    non_bonded_force: openmm.NonbondedForce,
    interchange: "Interchange",
    openff_openmm_particle_map: dict,
    parent_virtual_particle_mapping: defaultdict[int, list[int]],
):
    # The topology indices reported by toolkit methods must be converted to openmm indices
    bonds = [
        sorted(openff_openmm_particle_map[interchange.topology.atom_index(a)] for a in bond.atoms)
        for bond in interchange.topology.bonds
    ]

    coul_14 = getattr(data.electrostatics_collection, "scale_14", 1.0)
    vdw_14 = getattr(data.vdw_collection, "scale_14", 1.0)

    # First, create all atom-atom exceptions according to the conventional pattern
    non_bonded_force.createExceptionsFromBonds(
        bonds=bonds,
        coulomb14Scale=coul_14,
        lj14Scale=vdw_14,
    )

    # Faster to loop through exceptions and look up parents than opposite
    if parent_virtual_particle_mapping not in (None, dict()):
        for exception_index in range(non_bonded_force.getNumExceptions()):
            # These particles should only be atoms in this loop
            (
                p1,
                p2,
                _,
                _,
                _,
            ) = non_bonded_force.getExceptionParameters(exception_index)

            # Adding parent-(child of neighbor) exceptions from either p1 or p2
            for virtual_particle_of_p1 in parent_virtual_particle_mapping[p1]:
                if virtual_particle_of_p1 != p2:
                    _add_exception_from_existing_exception(
                        force=non_bonded_force,
                        existing_exception_index=exception_index,
                        new_p1=virtual_particle_of_p1,
                        new_p2=p2,
                        charge_scaling=coul_14,
                        vdw_scaling=vdw_14,
                    )

            # If this iterable is not empty, add an exception between p2's virtual
            # particle and the "other" atom in p2's exception
            for virtual_particle_of_p2 in parent_virtual_particle_mapping[p2]:
                if virtual_particle_of_p2 != p1:
                    _add_exception_from_existing_exception(
                        force=non_bonded_force,
                        existing_exception_index=exception_index,
                        new_p1=virtual_particle_of_p2,
                        new_p2=p1,
                        charge_scaling=coul_14,
                        vdw_scaling=vdw_14,
                    )

            # Adding (child of this parent)-(child of neighbor) exceptions
            for v1, v2 in itertools.product(
                parent_virtual_particle_mapping[p1],
                parent_virtual_particle_mapping[p2],
            ):
                _add_exception_from_existing_exception(
                    force=non_bonded_force,
                    existing_exception_index=exception_index,
                    new_p1=v1,
                    new_p2=v2,
                    charge_scaling=coul_14,
                    vdw_scaling=vdw_14,
                )

        for (
            parent,
            virtual_particles_of_this_parent,
        ) in parent_virtual_particle_mapping.items():
            for virtual_particle in virtual_particles_of_this_parent:
                # Add exceptions between each virtual particle and parent atom
                _add_zeroed_exception(non_bonded_force, parent, virtual_particle)

            # Add exceptions (all zeros) between each virtual site - virtual site pair
            # of a common parent such as the two dummy atoms on TIP5P
            for v1, v2 in itertools.product(
                virtual_particles_of_this_parent,
                virtual_particles_of_this_parent,
            ):
                # No need to add symmetric/duplicate exceptions, nor self-self exceptions
                if v1 < v2:
                    _add_zeroed_exception(non_bonded_force, v1, v2)


def _create_multiple_nonbonded_forces(
    data: _NonbondedData,
    interchange: "Interchange",
    system: openmm.System,
    ewald_tolerance: float,
    molecule_virtual_site_map: dict,
    openff_openmm_particle_map: dict[int | BaseVirtualSiteKey, int],
):
    from openff.interchange.components.toolkit import _get_14_pairs

    if molecule_virtual_site_map in (None, dict()):
        has_virtual_sites = False
    elif all([len(v) == 0 for v in molecule_virtual_site_map.values()]):
        has_virtual_sites = False
    else:
        has_virtual_sites = True

    vdw_force = _create_vdw_force(
        data,
        interchange,
        molecule_virtual_site_map,
        has_virtual_sites,
    )

    electrostatics_force: openmm.NonbondedForce = _create_electrostatics_force(
        data,
        interchange,
        ewald_tolerance,
        molecule_virtual_site_map,
        has_virtual_sites,
        openff_openmm_particle_map,
    )

    _set_particle_parameters(
        data,
        vdw_force,
        electrostatics_force,
        interchange,
        has_virtual_sites,
        molecule_virtual_site_map,
        openff_openmm_particle_map,
    )

    # Attempting to match the value used internally by OpenMM; The source of this value is likely
    # https://github.com/openmm/openmm/issues/1149#issuecomment-250299854
    # 1 / * (4pi * eps0) * elementary_charge ** 2 / nanometer ** 2
    coul_const = 138.935456  # kJ/nm

    if vdw_force is not None:
        vdw = data.vdw_collection

        if vdw.is_plugin:
            # TODO: Custom mixing rules in plugins is untested
            vdw_14_force = openmm.CustomBondForce(
                _get_scaled_potential_function(data.vdw_expression),  # type: ignore[arg-type]
            )
            vdw_14_force.setName("vdW 1-4 force")

            # feed in r_min1, epsilon1, ..., r_min2, epsilon2, ... each as individual parameters
            for index in [1, 2]:
                for parameter in vdw.potential_parameters():
                    vdw_14_force.addPerBondParameter(f"{parameter}{index!s}")

            vdw_14_force.addGlobalParameter("scale14", vdw.scale_14)

            for global_parameter in vdw.global_parameters():  # type: ignore[attr-defined]
                vdw_14_force.addGlobalParameter(
                    global_parameter,
                    getattr(vdw, global_parameter).m,
                )

            for term, value in data.vdw_collection.pre_computed_terms().items():  # type: ignore[attr-defined]
                vdw_14_force.addGlobalParameter(term, value)

        else:
            vdw_expression: str = data.vdw_expression  # type: ignore

            vdw_14_force = openmm.CustomBondForce(vdw_expression)
            vdw_14_force.setName("vdW 1-4 force")

            for parameter in vdw.potential_parameters():
                vdw_14_force.addPerBondParameter(f"{parameter}")

        vdw_14_force.setUsesPeriodicBoundaryConditions(interchange.box is not None)

    else:
        vdw_14_force = None

    coul_14_force = openmm.CustomBondForce(f"{coul_const}*qq/r")
    coul_14_force.setName("Electrostatics 1-4 force")
    coul_14_force.addPerBondParameter("qq")
    coul_14_force.setUsesPeriodicBoundaryConditions(interchange.box is not None)

    coul_14, vdw_14 = _get_14_scaling_factors(data)

    openmm_pairs = set()

    for atom1, atom2 in _get_14_pairs(interchange.topology):
        openff_indices = (
            interchange.topology.atom_index(atom1),
            interchange.topology.atom_index(atom2),
        )

        openmm_indices = (
            openff_openmm_particle_map[openff_indices[0]],
            openff_openmm_particle_map[openff_indices[1]],
        )

        openmm_pairs.add(openmm_indices)

    if electrostatics_force is not None:
        for i in range(electrostatics_force.getNumExceptions()):
            (p1, p2, _, _, _) = electrostatics_force.getExceptionParameters(i)

            if (p1, p2) in openmm_pairs or (p2, p1) in openmm_pairs:
                if vdw_force is not None:
                    if data.vdw_collection.is_plugin:
                        # Since we fed in in r_min1, epsilon1, ..., r_min2, epsilon2, ...
                        # each as individual parameters, we need to prepare a list of
                        # length 2 * len(potential_parameters) to this constructor
                        parameters1 = vdw_force.getParticleParameters(p1)
                        parameters2 = vdw_force.getParticleParameters(p2)
                        vdw_14_force.addBond(p1, p2, [*parameters1, *parameters2])

                    else:
                        # Look up the vdW parameters for each particle
                        sig1, eps1 = vdw_force.getParticleParameters(p1)
                        sig2, eps2 = vdw_force.getParticleParameters(p2)

                        # manually compute ...
                        if data.mixing_rule == "lorentz-berthelot":
                            sig_14 = (sig1 + sig2) * 0.5
                            eps_14 = (eps1 * eps2) ** 0.5 * vdw_14
                        elif data.mixing_rule == "geometric":
                            sig_14 = (sig1 * sig2) ** 0.5
                            eps_14 = (eps1 * eps2) ** 0.5 * vdw_14
                        else:
                            raise UnsupportedExportError(
                                f"Unsupported mixing rule: {data.mixing_rule}",
                            )

                        # ... and set the 1-4 interactions
                        vdw_14_force.addBond(p1, p2, [sig_14, eps_14])

                # Look up the partial charges for each particle
                q1 = electrostatics_force.getParticleParameters(p1)[0]
                q2 = electrostatics_force.getParticleParameters(p2)[0]

                # manually compute ...
                qq = q1 * q2 * coul_14

                # ... and set the 1-4 interactions
                coul_14_force.addBond(p1, p2, [qq])

            if vdw_force is not None:
                vdw_force.addExclusion(p1, p2)

            if electrostatics_force is not None:
                electrostatics_force.setExceptionParameters(i, p1, p2, 0.0, 0.0, 0.0)

    for force in [vdw_force, electrostatics_force, vdw_14_force, coul_14_force]:
        if force is not None:
            system.addForce(force)

    if vdw_force is not None and electrostatics_force is not None:
        vdw_uses = vdw_force.usesPeriodicBoundaryConditions()
        elec_uses = electrostatics_force.usesPeriodicBoundaryConditions()
        if vdw_uses != elec_uses:
            raise UnsupportedCutoffMethodError(
                "When using `openmm.CustomNonbondedForce`, vdW and electrostatics cutoff methods "
                "must agree on whether or not periodic boundary conditions should be used. "
                f"OpenMM will throw an error. Found vdw method {vdw_force.getNonbondedMethod()}, "
                f"and electrostatics method {electrostatics_force.getNonbondedMethod()}, ",
            )


def _create_vdw_force(
    data: _NonbondedData,
    interchange: "Interchange",
    molecule_virtual_site_map: dict[int, list[BaseVirtualSiteKey]],
    has_virtual_sites: bool,
) -> openmm.CustomNonbondedForce | None:
    vdw_collection: vdWCollection | None = data.vdw_collection

    if vdw_collection is None:
        return None

    vdw_expression: str = data.vdw_expression  # type: ignore[assignment]
    mixing_rule_expression: str = _MIXING_RULE_EXPRESSIONS.get(
        data.mixing_rule,
        "",
    )

    vdw_force = openmm.CustomNonbondedForce(
        (
            f"{vdw_expression}"
            if mixing_rule_expression in (None, "")
            else f"{vdw_expression}; {mixing_rule_expression}"
        ),
    )
    vdw_force.setName("vdW force")

    for potential_parameter in vdw_collection.potential_parameters():
        vdw_force.addPerParticleParameter(potential_parameter)

    if vdw_collection.is_plugin:
        # TODO: Move this block outside of the plugin conditional
        for global_parameter in vdw_collection.global_parameters():  # type: ignore[attr-defined]
            vdw_force.addGlobalParameter(
                global_parameter,
                getattr(vdw_collection, global_parameter).m,
            )

        for term, value in vdw_collection.pre_computed_terms().items():  # type: ignore[attr-defined]
            vdw_force.addGlobalParameter(term, value)

    for molecule in interchange.topology.molecules:
        for _ in molecule.atoms:
            vdw_force.addParticle(vdw_collection.default_parameter_values())

    for molecule in interchange.topology.molecules:
        if has_virtual_sites:
            molecule_index = interchange.topology.molecule_index(molecule)
            for _ in molecule_virtual_site_map[molecule_index]:
                vdw_force.addParticle(vdw_collection.default_parameter_values())

    if data.vdw_method == "cutoff":
        if interchange.box is None:
            vdw_force.setNonbondedMethod(openmm.NonbondedForce.CutoffNonPeriodic)
        else:
            vdw_force.setNonbondedMethod(openmm.NonbondedForce.CutoffPeriodic)
        vdw_force.setUseLongRangeCorrection(True)
        vdw_force.setCutoffDistance(to_openmm_quantity(data.vdw_cutoff))

        _apply_switching_function(vdw_collection, vdw_force)

    elif data.vdw_method == "no-cutoff":
        if interchange.box is None:
            vdw_force.setNonbondedMethod(openmm.NonbondedForce.NoCutoff)
        else:
            raise UnsupportedCutoffMethodError(
                "vdW method 'no-cutoff' is not valid for periodic systems.",
            )

    elif data.vdw_method == "pme":
        if interchange.box is None:
            raise UnsupportedCutoffMethodError(
                "vdW method pme/ljpme is not valid for non-periodic systems.",
            )
        else:
            raise UnsupportedCutoffMethodError(
                "LJ-PME with split non-bonded forces is not supported due to openmm.CustomNonbondedForce "
                "not supporting PME. If also using PME electrostatics, try `combine_nonbonded_forces=True`,  "
                "which should produce a single force with NonbondedForce.LJPME, which uses PME for both "
                "electrostatics and LJ forces terms. If your use case would benefit from split non-bonded "
                "forces with LJPME, please open a feature request.",
            )

    else:
        raise UnsupportedCutoffMethodError(
            f"Unsupported vdW method: {data.vdw_method}",
        )

    return vdw_force


def _create_electrostatics_force(
    data: _NonbondedData,
    interchange: "Interchange",
    ewald_tolerance: float,
    molecule_virtual_site_map: dict[int, list[BaseVirtualSiteKey]],
    has_virtual_sites: bool,
    openff_openmm_particle_map,
) -> openmm.NonbondedForce | None:
    if data.electrostatics_collection is None:
        return None

    electrostatics_force = openmm.NonbondedForce()
    electrostatics_force.setName("Electrostatics force")

    # mapping between (openmm) index of each atom and the (openmm) index of each virtual particle
    #   of that parent atom (if any)
    # if no virtual sites at all, this remains an empty dict
    parent_virtual_particle_mapping: defaultdict[int, list[int]] = defaultdict(list)

    for molecule in interchange.topology.molecules:
        for _ in molecule.atoms:
            electrostatics_force.addParticle(0.0, 1.0, 0.0)

    for molecule in interchange.topology.molecules:
        if has_virtual_sites:
            molecule_index = interchange.topology.molecule_index(molecule)
            for virtual_site_key in molecule_virtual_site_map[molecule_index]:
                force_index = electrostatics_force.addParticle(0.0, 1.0, 0.0)

                # this is an "OpenFF" index
                parent_atom_index = virtual_site_key.orientation_atom_indices[0]

                # this dict contains only "OpenMM" indices, so look through map
                parent_virtual_particle_mapping[openff_openmm_particle_map[parent_atom_index]].append(force_index)

    if data.electrostatics_method == "reaction-field":
        warnings.warn(
            "OpenMM implements reaction field electrostatics with CutoffPeriodic and CutoffNonperiodic. "
            "See http://docs.openmm.org/latest/userguide/theory/02_standard_forces.html?highlight="
            "reaction%20field#coulomb-interaction-with-cutoff",
            NonbondedSettingsWarning,
        )

        if data.periodic:
            electrostatics_force.setNonbondedMethod(
                openmm.NonbondedForce.CutoffPeriodic,
            )
        else:
            electrostatics_force.setNonbondedMethod(
                openmm.NonbondedForce.CutoffNonPeriodic,
            )

        electrostatics_force.setCutoffDistance(
            to_openmm_quantity(data.electrostatics_collection.cutoff),
        )

    elif data.electrostatics_method == _PME:
        electrostatics_force.setNonbondedMethod(openmm.NonbondedForce.PME)
        electrostatics_force.setEwaldErrorTolerance(ewald_tolerance)
        electrostatics_force.setUseDispersionCorrection(True)
        if data.vdw_cutoff is not None:
            # All nonbonded forces must use the same cutoff, even though PME doesn't have a cutoff
            electrostatics_force.setCutoffDistance(
                to_openmm_quantity(data.vdw_cutoff),
            )
    elif data.electrostatics_method == "Coulomb":
        if interchange.box is None:
            electrostatics_force.setNonbondedMethod(openmm.NonbondedForce.NoCutoff)
        else:
            raise UnsupportedCutoffMethodError(
                f"Electrostatics method {data.electrostatics_method} ambiguous with a periodic system.",
            )

    elif data.electrostatics_method == "cutoff":
        raise UnsupportedCutoffMethodError(
            "OpenMM does not support electrostatics with a hard cutoff (and no reaction field "
            'modification). Consider using `"reaction-field"` to get force(s) with CutoffPeriodic '
            "or CutoffNonperiodic.",
        )

    else:
        raise UnsupportedCutoffMethodError(
            f"Electrostatics method {data.electrostatics_method} not supported",
        )

    _create_exceptions(
        data,
        electrostatics_force,
        interchange,
        openff_openmm_particle_map,
        parent_virtual_particle_mapping,
    )

    return electrostatics_force


def _set_particle_parameters(
    data: _NonbondedData,
    vdw_force: openmm.CustomNonbondedForce,
    electrostatics_force: openmm.NonbondedForce,
    interchange: "Interchange",
    has_virtual_sites: bool,
    molecule_virtual_site_map: dict[int, list[BaseVirtualSiteKey]],
    openff_openmm_particle_map: dict[int | BaseVirtualSiteKey, int],
):
    # TODO: Some funky plugins might have no charges, so there eventually should be
    #       handling for electrostatics_force = None
    electrostatics: ElectrostaticsCollection = data.electrostatics_collection

    partial_charges = _simpler_charges(electrostatics)

    vdw: vdWCollection = data.vdw_collection

    for molecule in interchange.topology.molecules:
        for atom in molecule.atoms:
            atom_index = interchange.topology.atom_index(atom)
            particle_index = openff_openmm_particle_map[atom_index]
            # TODO: Actually process virtual site vdW parameters here

            top_key = TopologyKey(atom_indices=(atom_index,))

            partial_charge = partial_charges[top_key].m_as("elementary_charge")

            if vdw is not None:
                pot_key = vdw.key_map[top_key]

                if vdw.is_plugin:
                    parameters = vdw.potentials[pot_key].parameters

                    if hasattr(vdw, "modify_parameters"):
                        # This method stripsopenmm.units ..
                        parameters = vdw.modify_parameters(parameters)
                    else:
                        # so manually strip them if the method is not present
                        parameters = {key: val.m for key, val in parameters.items()}

                else:
                    sigma = vdw.potentials[pot_key].parameters["sigma"]
                    epsilon = vdw.potentials[pot_key].parameters["epsilon"]

                    sigma = sigma.m_as("nanometer")
                    epsilon = epsilon.m_as("kilojoule / mole")
            else:
                sigma = openmm.unit.Quantity(0.0, openmm.unit.nanometer)
                epsilon = openmm.unit.Quantity(0.0, openmm.unit.kilojoules_per_mole)

            if vdw_force is not None:
                if vdw.is_plugin:
                    vdw_force.setParticleParameters(
                        particle_index,
                        [*parameters.values()],
                    )
                    del parameters
                else:
                    vdw_force.setParticleParameters(particle_index, [sigma, epsilon])

            if electrostatics_force is not None:
                electrostatics_force.setParticleParameters(
                    particle_index,
                    partial_charge,
                    0.0,
                    0.0,
                )

        for virtual_site_key in molecule_virtual_site_map[interchange.topology.molecule_index(molecule)]:
            particle_index = openff_openmm_particle_map[virtual_site_key]

            if vdw is not None:
                pot_key = vdw.key_map[virtual_site_key]

                if vdw.is_plugin:
                    parameters = vdw.potentials[pot_key].parameters

                    if hasattr(vdw, "modify_parameters"):
                        # This method strips openmm.units ..

                        parameters = vdw.modify_parameters(parameters)

                    else:
                        # so manually strip them if the method is not present
                        parameters = {key: val.m for key, val in parameters.items()}

                    # a non-LJ vdW interaction might be mixed with virtual site parameters that have
                    # zeroed-out sigma and epsilon; in this case
                    if {tuple(vdw.potentials[pot_key].parameters.keys())} != {
                        vdw.potential_parameters(),
                    }:
                        parameters = {
                            key: val
                            for key, val in zip(
                                vdw.potential_parameters(),
                                vdw.default_parameter_values(),
                            )
                        }

                else:
                    sigma = vdw.potentials[pot_key].parameters["sigma"]
                    epsilon = vdw.potentials[pot_key].parameters["epsilon"]

                    sigma = sigma.m_as("nanometer")
                    epsilon = epsilon.m_as("kilojoule / mole")
            else:
                sigma = openmm.unit.Quantity(0.0, openmm.unit.nanometer)
                epsilon = openmm.unit.Quantity(0.0, openmm.unit.kilojoules_per_mole)

            if vdw_force is not None:
                # TODO: Refactor getting the "vdW" parameters by particle index into
                #       a common function

                if vdw.is_plugin:
                    vdw_force.setParticleParameters(
                        particle_index,
                        [*parameters.values()],
                    )
                else:
                    vdw_force.setParticleParameters(particle_index, [sigma, epsilon])

            partial_charge = partial_charges[virtual_site_key].m_as("elementary_charge")

            if electrostatics_force is not None:
                electrostatics_force.setParticleParameters(
                    particle_index,
                    partial_charges[virtual_site_key].m_as("elementary_charge"),
                    0.0,
                    0.0,
                )


def _get_14_scaling_factors(data: _NonbondedData) -> tuple[float, float]:
    if data.electrostatics_collection is None:
        coul_14 = 1.0
    else:
        coul_14 = data.electrostatics_collection.scale_14

    if data.vdw_collection is None:
        vdw_14 = 1.0
    else:
        vdw_14 = data.vdw_collection.scale_14

    return coul_14, vdw_14


def _apply_switching_function(
    vdw_collection: "vdWCollection",
    force: openmm.NonbondedForce,
):
    if not hasattr(force, "setUseSwitchingFunction"):
        raise CannotSetSwitchingFunctionError(
            "Attempting to set switching function on an OpenMM force that does nont support it."
            f"Passed force of type {type(force)}.",
        )
    if getattr(vdw_collection, "switch_width", None) is None:
        force.setUseSwitchingFunction(False)
    elif vdw_collection.switch_width.m == 0.0:
        force.setUseSwitchingFunction(False)
    else:
        switching_distance = to_openmm_quantity(
            vdw_collection.cutoff - vdw_collection.switch_width,
        )

        if switching_distance._value < 0:
            raise UnsupportedCutoffMethodError(
                "Found a `switch_width` greater than the cutoff distance. It's not clear "
                "what this means and it's probably invalid. Found "
                f"switch_width{vdw_collection.switch_width} and cutoff {vdw_collection.cutoff}",
            )

        force.setUseSwitchingFunction(True)
        force.setSwitchingDistance(switching_distance)


def _get_scaled_potential_function(potential: str) -> str:
    # https://github.com/openforcefield/smirnoff-plugins/blob/979907fc93e8b6217d40e6f469809c230b86b012/smirnoff_plugins/handlers/nonbonded.py#L63-L83
    split_potential = potential.split(";")

    for i, expression in enumerate(split_potential):
        if "=" not in expression:
            # This is the actual expression, final energy so modify
            split_potential[i] = f"({expression})*scale14"
            break

    return ";".join(split_potential)


def _add_exception_from_existing_exception(
    force: openmm.NonbondedForce,
    existing_exception_index: int,
    new_p1: int,
    new_p2: int,
    charge_scaling: float = 0.8333333333,
    vdw_scaling: float = 0.5,
):
    """Map the interactions of an existing exception onto a new particle pair."""
    _, _, charge_product, _, epsilon = force.getExceptionParameters(
        existing_exception_index,
    )

    # Assume this means that the existing p1-p2 pair is a 1-2 or 1-3 interaction
    if charge_product._value == epsilon._value == 0.0:
        _add_zeroed_exception(force, new_p1, new_p2)

    else:
        # TODO: Using mixing rule here, do not simply assume Lorentz-Berthelot
        p1_parameters = force.getParticleParameters(new_p1)
        p2_parameters = force.getParticleParameters(new_p2)

        force.addException(
            particle1=new_p1,
            particle2=new_p2,
            chargeProd=p1_parameters[0] * p2_parameters[0] * charge_scaling,
            sigma=(p1_parameters[1] + p2_parameters[1]) * 0.5,
            epsilon=(p1_parameters[2] * p2_parameters[2]) ** 0.5 * vdw_scaling,
        )


def _add_zeroed_exception(
    force: openmm.NonbondedForce,
    particle1: int,
    particle2: int,
):
    force.addException(
        particle1=particle1,
        particle2=particle2,
        chargeProd=0.0,
        sigma=1.0,
        epsilon=0.0,
        replace=True,
    )



"""Functions for running energy evluations with molecular simulation engines."""

from openff.interchange.drivers.all import get_all_energies, get_summary_data
from openff.interchange.drivers.amber import get_amber_energies
from openff.interchange.drivers.gromacs import get_gromacs_energies
from openff.interchange.drivers.lammps import get_lammps_energies
from openff.interchange.drivers.openmm import get_openmm_energies
from openff.interchange.drivers.report import EnergyReport

__all__ = [
    "EnergyReport",
    "get_all_energies",
    "get_amber_energies",
    "get_gromacs_energies",
    "get_lammps_energies",
    "get_openmm_energies",
    "get_summary_data",
]



"""Functions for running energy evluations with GROMACS."""

import subprocess
import tempfile
from importlib import resources
from pathlib import Path
from shutil import which

from openff.toolkit import Quantity
from openff.utilities.utilities import requires_package, temporary_cd

from openff.interchange import Interchange
from openff.interchange.components.mdconfig import MDConfig
from openff.interchange.drivers.report import EnergyReport
from openff.interchange.exceptions import (
    GMXGromppError,
    GMXMdrunError,
    GMXNotFoundError,
)


def _find_gromacs_executable(raise_exception: bool = False) -> str | None:
    """Attempt to locate a GROMACS executable based on commonly-used names."""
    gromacs_executable_names = ["gmx", "gmx_mpi", "gmx_d", "gmx_mpi_d"]

    for name in gromacs_executable_names:
        if which(name):
            return name

    if raise_exception:
        raise GMXNotFoundError
    else:
        return None


def _get_mdp_file(key: str = "auto") -> str:
    mapping = {
        "default": "default.mdp",
        "cutoff": "cutoff.mdp",
        "cutoff_hbonds": "cutoff_hbonds.mdp",
        "cutoff_buck": "cutoff_buck.mdp",
    }

    dir_path = resources.files("openff.interchange._tests.data.mdp")
    return str(dir_path / mapping[key])


def get_gromacs_energies(
    interchange: Interchange,
    mdp: str = "auto",
    round_positions: int = 8,
    detailed: bool = False,
    _merge_atom_types: bool = False,
    _monolithic: bool = True,
) -> EnergyReport:
    """
    Given an OpenFF Interchange object, return single-point energies as computed by GROMACS.

    Parameters
    ----------
    interchange
        An OpenFF Interchange object to compute the single-point energy of
    mdp
        A string key identifying the GROMACS `.mdp` file to be used. See `_get_mdp_file`.
    round_positions
        A decimal precision for the positions in the `.gro` file.
    detailed
        If True, return a detailed report containing the energies of each term.
    _merge_atom_types
        If True, energy should be computed with merging atom types.

    Returns
    -------
    report
        An `EnergyReport` object containing the single-point energies.

    """
    return _process(
        _get_gromacs_energies(
            interchange=interchange,
            mdp=mdp,
            round_positions=round_positions,
            merge_atom_types=_merge_atom_types,
            monolithic=_monolithic,
        ),
        detailed=detailed,
    )


def _get_gromacs_energies(
    interchange: Interchange,
    mdp: str = "auto",
    round_positions: int = 8,
    merge_atom_types: bool = False,
    monolithic: bool = True,
) -> dict[str, Quantity]:
    with tempfile.TemporaryDirectory() as tmpdir:
        with temporary_cd(tmpdir):
            prefix = "_tmp"
            interchange.to_gromacs(
                prefix=prefix,
                decimal=round_positions,
                monolithic=monolithic,
                _merge_atom_types=merge_atom_types,
            )

            if mdp == "auto":
                mdconfig = MDConfig.from_interchange(interchange)
                mdp_file = "tmp.mdp"
                mdconfig.write_mdp_file(mdp_file)
            else:
                mdp_file = _get_mdp_file(mdp)

            return _run_gmx_energy(
                top_file="_tmp.top",
                gro_file="_tmp.gro",
                mdp_file=mdp_file,
                maxwarn=2,
            )


def _run_gmx_energy(
    top_file: Path | str,
    gro_file: Path | str,
    mdp_file: Path | str,
    maxwarn: int = 1,
) -> dict[str, Quantity]:
    """
    Given GROMACS files, return single-point energies as computed by GROMACS.

    Parameters
    ----------
    top_file
        The path to a GROMACS topology (`.top`) file.
    gro_file
        The path to a GROMACS coordinate (`.gro`) file.
    mdp_file
        The path to a GROMACS molecular dynamics parameters (`.mdp`) file.
    maxwarn
        The number of warnings to allow when `gmx grompp` is called (via the `-maxwarn` flag).

    Returns
    -------
    energies
        A dictionary of energies, keyed by the GROMACS energy term name.

    """
    gmx = _find_gromacs_executable(raise_exception=True)

    grompp_cmd = f"{gmx} grompp --maxwarn {maxwarn} -o out.tpr"
    grompp_cmd += f" -f {mdp_file} -c {gro_file} -p {top_file}"

    grompp = subprocess.Popen(
        grompp_cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    _, err = grompp.communicate()

    if grompp.returncode:
        raise GMXGromppError(err)

    # Some GROMACS builds will want `-ntmpi` instead of `ntomp`
    mdrun_cmd = f"{gmx} mdrun -s out.tpr -e out.edr -ntomp 1"

    mdrun = subprocess.Popen(
        mdrun_cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    _, err = mdrun.communicate()

    if mdrun.returncode:
        raise GMXMdrunError(err)

    return _parse_gmx_energy("out.edr")


def _get_gmx_energy_vdw(gmx_energies: dict) -> Quantity:
    """Get the total nonbonded energy from a set of GROMACS energies."""
    gmx_vdw = Quantity(0.0, "kilojoule / mole")
    for key in ["LJ (SR)", "LJ recip.", "LJ-14", "Disper. corr.", "Buck.ham (SR)"]:
        if key in gmx_energies:
            gmx_vdw += gmx_energies[key]

    return gmx_vdw


def _get_gmx_energy_coul(gmx_energies: dict) -> Quantity:
    gmx_coul = Quantity(0.0, "kilojoule / mole")
    for key in ["Coulomb (SR)", "Coul. recip.", "Coulomb-14"]:
        if key in gmx_energies:
            gmx_coul += gmx_energies[key]

    return gmx_coul


def _get_gmx_energy_torsion(gmx_energies: dict) -> Quantity:
    """Canonicalize torsion energies from a set of GROMACS energies."""
    gmx_torsion = Quantity(0.0, "kilojoule / mole")

    for key in ["Torsion", "Proper Dih.", "Per. Imp. Dih."]:
        if key in gmx_energies:
            gmx_torsion += gmx_energies[key]

    return gmx_torsion


@requires_package("pyedr")
def _parse_gmx_energy(edr_path: str) -> dict[str, Quantity]:
    """Parse an `.edr` file written by `gmx energy`."""
    from pyedr import read_edr

    keys_to_drop = [
        "Time",
        "Kinetic En.",
        "Temperature",
        "Pres. DC",
        "Pressure",
        "Vir-XX",
        "Vir-YY",
        "Vir-ZZ",
        "Vir-YX",
        "Vir-XY",
        "Vir-YZ",
        "Vir-XZ",
    ]

    all_energies, all_names, _ = read_edr(edr_path, verbose=False)
    parsed_energies = {}

    for idx, name in enumerate(all_names):
        if name not in keys_to_drop:
            parsed_energies[name] = all_energies[0][idx]

    return {key: Quantity(val, "kilojoule / mole") for key, val in parsed_energies.items()}


def _process(
    energies: dict[str, Quantity],
    detailed: bool = False,
) -> EnergyReport:
    """Process energies from GROMACS into a standardized format."""
    if detailed:
        return EnergyReport(energies=_canonicalize_detailed_energies(energies))

    return EnergyReport(
        energies={
            "Bond": energies.get("Bond", Quantity(0.0, "kilojoule / mole")),
            "Angle": energies.get("Angle", Quantity(0.0, "kilojoule / mole")),
            "Torsion": _get_gmx_energy_torsion(energies),
            "RBTorsion": energies.get("Ryckaert-Bell.", Quantity(0.0, "kilojoule / mole")),
            "vdW": _get_gmx_energy_vdw(energies),
            "Electrostatics": _get_gmx_energy_coul(energies),
        },
    )


def _canonicalize_detailed_energies(
    energies: dict[str, Quantity],
) -> dict[str, Quantity]:
    """Condense the full `gmx energy` report into the keys of a "detailed" report."""
    return {
        "Bond": energies.get("Bond", Quantity(0.0, "kilojoule / mole")),
        "Angle": energies.get("Angle", Quantity(0.0, "kilojoule / mole")),
        "Torsion": _get_gmx_energy_torsion(energies),
        "vdW": energies.get("LJ (SR)", Quantity(0.0, "kilojoule / mole"))
        + energies.get("LJ recip.", Quantity(0.0, "kilojoule / mole")),
        "vdW 1-4": energies.get("LJ-14", Quantity(0.0, "kilojoule / mole")),
        "Electrostatics": energies.get("Coulomb (SR)", Quantity(0.0, "kilojoule / mole"))
        + energies.get(
            "Coul. recip.",
            Quantity(0.0, "kilojoule / mole"),
        ),
        "Electrostatics 1-4": energies.get("Coulomb-14", Quantity(0.0, "kilojoule / mole")),
    }



import re

from openff.interchange.models import (
    AngleKey,
    BaseVirtualSiteKey,
    BondKey,
    ImproperTorsionKey,
    PotentialKey,
    ProperTorsionKey,
    SMIRNOFFVirtualSiteKey,
    TopologyKey,
    VirtualSiteKey,
)


def test_potentialkey_hash_uniqueness():
    """Test that PotentialKey hashes differ when optional attributes are set."""

    smirks = "[#1:1]-[#8X2:2]"
    ref = PotentialKey(id=smirks)
    with_mult = PotentialKey(id=smirks, mult=2)
    with_associated_handler = PotentialKey(id=smirks, associated_handler="espaloma")
    with_bond_order = PotentialKey(id=smirks, bond_order=5 / 4)

    keys = [ref, with_mult, with_associated_handler, with_bond_order]
    assert len({hash(k) for k in keys}) == len(keys)


def test_topologykey_hash_uniqueness():
    """Test that TopologyKey hashes differ when optional attributes are set."""

    smirks = "[#1:1]-[#8X2:2]"
    ref = ProperTorsionKey(id=smirks, atom_indices=(2, 0, 1, 3))
    with_mult = ProperTorsionKey(id=smirks, atom_indices=(2, 0, 1, 3), mult=2)
    with_bond_order = ProperTorsionKey(
        id=smirks,
        atom_indices=(2, 0, 1, 3),
        bond_order=1.4,
    )

    keys = [ref, with_mult, with_bond_order]
    assert len({hash(k) for k in keys}) == len(keys)


def test_virtualsitekey_hash_uniqueness():
    """Test that VirtualSiteKey hashes differ when optional attributes are set."""

    ref = VirtualSiteKey(
        orientation_atom_indices=(0, 1),
        name="vs1",
        type="BondCharge",
        match="once",
    )
    with_name = VirtualSiteKey(
        orientation_atom_indices=(0, 1),
        name="vs2",
        type="TrivalentLonePair",
        match="once",
    )
    with_type = VirtualSiteKey(
        orientation_atom_indices=(0, 1),
        name="vs1",
        type="TrivalentLonePair",
        match="once",
    )
    with_match = VirtualSiteKey(
        orientation_atom_indices=(0, 1),
        name="vs1",
        type="BondCharge",
        match="all_permutations",
    )

    keys = [ref, with_type, with_match, with_name]
    assert len({hash(k) for k in keys}) == len(keys)


def test_central_atom_improper():
    key = ImproperTorsionKey(id="foo", atom_indices=(0, 2, 1, 3))

    assert key.get_central_atom_index() == 0


def test_reprs():
    topology_key = TopologyKey(atom_indices=(10,))

    assert "atom indices (10,)" in repr(topology_key)
    assert "bond order" not in repr(topology_key)
    assert "mult" not in repr(topology_key)

    bond_key = BondKey(atom_indices=(3, 4))

    assert "atom indices (3, 4)" in repr(bond_key)
    assert "bond order" not in repr(bond_key)
    assert "mult" not in repr(bond_key)

    torsion_key = ProperTorsionKey(
        atom_indices=(0, 1, 10, 11),
        mult=2,
        bond_order=1.111,
    )

    assert "atom indices (0, 1, 10, 11)" in repr(torsion_key)
    assert "mult 2" in repr(torsion_key)
    assert "bond order 1.111" in repr(torsion_key)

    potential_key = PotentialKey(id="foobar")

    assert "foobar" in repr(potential_key)
    assert "mult" not in repr(potential_key)
    assert "bond order" not in repr(potential_key)

    potential_key = PotentialKey(id="blah", mult=2, bond_order=1.111)

    assert "blah" in repr(potential_key)
    assert "mult 2" in repr(potential_key)
    assert "bond order 1.111" in repr(potential_key)


def test_bondkey_eq_hash():
    """
    When __eq__ is true, the hashes must be equal.

    The converse is not required in Python; hash collisions between unequal
    objects are allowed and will be handled according to __eq__, possibly
    with a small runtime cost.
    """

    assert BondKey(atom_indices=(1, 3)) == BondKey(atom_indices=(1, 3))
    assert BondKey(atom_indices=(1, 3)) == TopologyKey(atom_indices=(1, 3))
    assert BondKey(atom_indices=(1, 3)) == (1, 3)
    assert hash(BondKey(atom_indices=(1, 3))) == hash((1, 3))
    assert BondKey(atom_indices=(1, 3)) != (1, 4)
    assert BondKey(atom_indices=(1, 3)) != (3, 1)
    assert BondKey(atom_indices=(1, 3), bond_order=None) == (1, 3)
    assert hash(BondKey(atom_indices=(1, 3), bond_order=None)) == hash((1, 3))
    assert BondKey(atom_indices=(1, 3), bond_order=None) != ((1, 3), None)
    assert BondKey(atom_indices=(1, 3), bond_order=1.5) != (1, 3)
    assert BondKey(atom_indices=(1, 3), bond_order=1.5) != ((1, 3), None)
    assert BondKey(atom_indices=(1, 3), bond_order=1.5) == ((1, 3), 1.5)


def test_anglekey_eq_hash():
    """
    When __eq__ is true, the hashes must be equal.

    The converse is not required in Python; hash collisions between unequal
    objects are allowed and will be handled according to __eq__, possibly
    with a small runtime cost.
    """
    assert AngleKey(atom_indices=(1, 3, 16)) == AngleKey(atom_indices=(1, 3, 16))
    assert AngleKey(atom_indices=(1, 3, 16)) == TopologyKey(atom_indices=(1, 3, 16))
    assert AngleKey(atom_indices=(1, 3, 16)) == (1, 3, 16)
    assert hash(AngleKey(atom_indices=(1, 3, 16))) == hash((1, 3, 16))
    assert AngleKey(atom_indices=(1, 3, 16)) != (16, 3, 1)
    assert AngleKey(atom_indices=(1, 3, 16)) != (1, 3)
    assert AngleKey(atom_indices=(1, 3, 16)) != (1, 3, 15)


def test_torsionkey_eq_hash():
    """
    When __eq__ is true, the hashes must be equal.

    The converse is not required in Python; hash collisions between unequal
    objects are allowed and will be handled according to __eq__, possibly
    with a small runtime cost.
    """
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4)) == ProperTorsionKey(
        atom_indices=(1, 2, 3, 4),
    )
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4)) == TopologyKey(
        atom_indices=(1, 2, 3, 4),
    )
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4)) == (1, 2, 3, 4)
    assert hash(ProperTorsionKey(atom_indices=(1, 2, 3, 4))) == hash((1, 2, 3, 4))
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4)) != (4, 3, 2, 1)
    assert ProperTorsionKey(
        atom_indices=(1, 2, 3, 4),
        mult=None,
        phase=None,
        bond_order=None,
    ) == (1, 2, 3, 4)
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), mult=0) != (1, 2, 3, 4)
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), phase=1.5) != (1, 2, 3, 4)
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), bond_order=1.5) != (1, 2, 3, 4)

    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), mult=0) == (
        (1, 2, 3, 4),
        0,
        None,
        None,
    )
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), phase=1.5) == (
        (1, 2, 3, 4),
        None,
        1.5,
        None,
    )
    assert ProperTorsionKey(atom_indices=(1, 2, 3, 4), bond_order=1.5) == (
        (1, 2, 3, 4),
        None,
        None,
        1.5,
    )


def test_base_virtual_site_eq():
    key1 = BaseVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="A", name="B")
    key2 = BaseVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="F", name="B")
    key3 = BaseVirtualSiteKey(orientation_atom_indices=[3, 2, 1], type="A", name="B")
    key4 = BaseVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="A", name="B")

    for key in [key1, key2, key3, key4]:
        assert key == key

    assert key1 != key2
    assert key1 != key3
    assert key2 != key3
    assert key1 == key4

    assert hash(key1) != hash(key2)
    assert hash(key1) != hash(key3)
    assert hash(key2) != hash(key3)
    assert hash(key1) == hash(key4)


def test_smirnoff_virtual_site_eq():
    """This class is the same as BaseVirtualSiteKey but adds the `match` field."""
    key1 = SMIRNOFFVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="A", name="B", match="once")
    key2 = SMIRNOFFVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="A", name="B", match="all_permutations")
    key3 = SMIRNOFFVirtualSiteKey(orientation_atom_indices=[1, 2, 3], type="A", name="B", match="once")

    for key in [key1, key2, key3]:
        assert key == key

    assert key1 != key2
    assert key1 == key3
    assert key2 != key3

    assert hash(key1) != hash(key2)
    assert hash(key1) == hash(key3)
    assert hash(key2) != hash(key3)


def test_virtual_site_key_repr():
    key = BaseVirtualSiteKey(orientation_atom_indices=[3, 4, 5], type="fOO", name="bAR")

    assert re.search("orientation atom ind.*3, 4, 5", repr(key))



import numpy
from openff.toolkit import Quantity, Topology

from openff.interchange._tests import _rng
from openff.interchange.common._positions import _infer_positions


class TestInferPositions:
    def test_short_circuit(self, methane):
        positions = Quantity(
            _rng.random((methane.n_atoms, 3)),
            "angstrom",
        )

        assert numpy.array_equal(
            _infer_positions(methane.to_topology(), positions).m_as("angstrom"),
            positions.m_as("angstrom"),
        )

    def test_basic(self, methane_with_conformer):
        assert numpy.array_equal(
            _infer_positions(methane_with_conformer.to_topology(), None).m_as(
                "angstrom",
            ),
            methane_with_conformer.conformers[0].m_as("angstrom"),
        )

    def test_multimolecule(self, methane_with_conformer, ethanol_with_conformer):
        topology = Topology.from_molecules(
            [methane_with_conformer, ethanol_with_conformer],
        )
        expected_positions = Quantity(
            numpy.concatenate(
                [
                    methane_with_conformer.conformers[0].m,
                    ethanol_with_conformer.conformers[0].m,
                ],
            ),
            "angstrom",
        )

        assert numpy.array_equal(
            _infer_positions(topology).m_as("angstrom"),
            expected_positions.m_as("angstrom"),
        )

    def test_mixed_conformers(self, methane_with_conformer, ethanol):
        topology = Topology.from_molecules([methane_with_conformer, ethanol])
        assert _infer_positions(topology) is None



"""Tests reproducing specific issues that are otherwise uncategorized."""

import random
import warnings

import numpy
import pytest
from openff.packmol import UNIT_CUBE, pack_box, solvate_topology
from openff.toolkit import ForceField, Molecule, Quantity, Topology
from openff.utilities import get_data_file_path, has_executable, skip_if_missing

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, _rng, needs_gmx, shuffle_topology
from openff.interchange.drivers import get_amber_energies, get_gromacs_energies, get_openmm_energies
from openff.interchange.exceptions import NonperiodicNoCutoffNotSupportedError
from openff.interchange.warnings import PresetChargesAndVirtualSitesWarning


def test_issue_723():
    parmed = pytest.importorskip("parmed")
    force_field = ForceField("openff-2.1.0.offxml")

    molecule = Molecule.from_smiles("C#N")
    molecule.generate_conformers(n_conformers=1)

    force_field.create_interchange(molecule.to_topology()).to_top("_x.top")

    parmed.load_file("_x.top")


@pytest.mark.skipif(not has_executable("packmol"), reason="Packmol is not installed")
@pytest.mark.parametrize("pack", [True, False])
def test_issue_1022(pack):
    topology = Topology.from_molecules(
        [
            MoleculeWithConformer.from_smiles(smi)
            for smi in [
                "CBr",
                "O",
                "O",
                "O",
                "[Na+]",
                "[Cl-]",
            ]
        ],
    )

    topology.box_vectors = Quantity(numpy.eye(3) * 10.0, "nanometer")

    if pack:
        topology = pack_box(
            molecules=[*topology.unique_molecules],
            number_of_copies=[1, 3, 1, 1],
            box_vectors=topology.box_vectors,
        )

    force_field = ForceField(
        "openff-2.0.0.offxml",
        get_data_file_path(
            "example-sigma-hole-bromine.offxml",
            "openff.interchange._tests.data",
        ),
    )

    interchange = force_field.create_interchange(topology)

    interchange.to_top("tmp")

    if pack:
        for seed in random.sample(range(0, 10**10), 5):
            # TODO: Compare GROMACS energies here as well
            get_openmm_energies(interchange).compare(
                get_openmm_energies(
                    shuffle_topology(
                        interchange,
                        force_field,
                        seed=seed,
                    ),
                ),
                tolerances={"Nonbonded": Quantity("1e-3 kilojoule_per_mole")},
            )


@skip_if_missing("openmm")
def test_issue_1031():
    import openmm.app

    # just grab some small PDB file from the toolkit, doesn't need to be huge, just
    # needs to include some relevant atom names
    openmm_topology = openmm.app.PDBFile(
        get_data_file_path(
            "proteins/MainChain_HIE.pdb",
            "openff.toolkit",
        ),
    ).topology

    openmm_atom_names = {atom.name for atom in openmm_topology.atoms()}

    interchange = Interchange.from_openmm(
        system=openmm.app.ForceField(
            "amber99sb.xml",
            "tip3p.xml",
        ).createSystem(
            openmm_topology,
            nonbondedMethod=openmm.app.PME,
        ),
        topology=openmm_topology,
    )

    openff_atom_names = {atom.name for atom in interchange.topology.atoms}

    assert sorted(openmm_atom_names) == sorted(openff_atom_names)

    # check a few atom names to ensure these didn't end up being empty sets
    for atom_name in ("NE2", "H3", "HA", "CH3", "CA", "CB", "CE1"):
        assert atom_name in openff_atom_names


def test_issue_1049():
    pytest.importorskip("openmm")

    topology = Topology.from_molecules(
        [
            Molecule.from_smiles("C"),
            Molecule.from_smiles("O"),
            Molecule.from_smiles("O"),
        ],
    )

    interchange = ForceField("openff-2.2.0.offxml", "opc.offxml").create_interchange(topology)

    openmm_topology = interchange.to_openmm_topology()
    openmm_system = interchange.to_openmm_system()

    # the same index in system should also be a virtual site in the topology
    for particle_index, particle in enumerate(openmm_topology.atoms()):
        assert openmm_system.isVirtualSite(particle_index) == (particle.element is None), (
            f"particle index {particle_index} is a virtual site in the system OR topology but not both"
        )


def test_issue_1052(sage, ethanol):
    """Test that _SMIRNOFFElectrostaticsCollection.charges is populated."""
    out = sage.create_interchange(ethanol.to_topology())

    assert len(out["Electrostatics"].charges) > 0


@skip_if_missing("openmm")
def test_issue_1209(sage, ethanol):
    ethanol.assign_partial_charges("gasteiger")
    ethanol.generate_conformers(n_conformers=1)

    out = sage.create_interchange(ethanol.to_topology(), charge_from_molecules=[ethanol])

    get_openmm_energies(out, combine_nonbonded_forces=False).compare(
        get_openmm_energies(
            Interchange.model_validate_json(out.model_dump_json()),
            combine_nonbonded_forces=False,
        ),
    )


def test_issue_1337(water):
    """Reproduce Issue #1337."""
    ff14sb_tip3p = ForceField("ff14sb_off_impropers_0.0.3.offxml", "tip3p.offxml")

    # just make sure this doesn't crash (original issue)
    # more extensive tests in openff/interchange/_tests/unit_tests/interop/openmm/test_constraints.py
    ff14sb_tip3p.create_interchange(water.to_topology())


@needs_gmx
def test_issue_1361_gromacs(caffeine, sage, tmp_path):
    """Test that the 'how to opt in to pseudo-vacuum' message is communicated."""
    interchange = sage.create_interchange(caffeine.to_topology())

    with pytest.raises(
        NonperiodicNoCutoffNotSupportedError,
        match=r"GROMACS versions 2020 and newer do not support systems without periodicity",
    ):
        interchange.to_gromacs(prefix="foo")


@pytest.mark.skipif(not has_executable("sander"), reason="sander not installed")
def test_issue_1361_amber(caffeine, sage, tmp_path):
    """Test that the 'how to opt in to pseudo-vacuum' message is communicated."""
    interchange = sage.create_interchange(caffeine.to_topology())

    with pytest.raises(
        NonperiodicNoCutoffNotSupportedError,
        match=r"vdW method no-cutoff not supported",
    ):
        interchange.to_amber(prefix="bar")


@pytest.mark.skipif(not has_executable("sander"), reason="sander not installed")
def test_issue_1395_amber(caffeine, sage, water, tmp_path):
    """Test that water charges are correctly ordered in ligand + water systems to Amber."""

    # turn off switching to make energy comparisons easier
    sage["vdW"].switch_width = Quantity(0.0, "nanometer")

    parmed = pytest.importorskip("parmed")

    topology = solvate_topology(
        topology=caffeine.to_topology(),
        nacl_conc=Quantity("0.0 mole/liter"),
        padding=Quantity("1.0 nanometer"),
        box_shape=UNIT_CUBE,
        target_density=Quantity("0.7 gram / centimeter**3"),
    )

    interchange = sage.create_interchange(topology)

    interchange.to_amber("solvated_caffeine")

    structure = parmed.load_file("solvated_caffeine.prmtop")

    # these should all be oxygens
    assert set([a.charge for a in structure.atoms][caffeine.n_atoms :: 3]) == {-0.834}

    # these should all be hydrogens
    assert set([a.charge for a in structure.atoms][caffeine.n_atoms + 1 :: 3]) == {0.417}
    assert set([a.charge for a in structure.atoms][caffeine.n_atoms + 2 :: 3]) == {0.417}

    get_openmm_energies(interchange, combine_nonbonded_forces=False).compare(
        get_amber_energies(interchange),
        tolerances={
            "Angle": Quantity("1e10 kilojoule_per_mole"),  # constraints weirdness
            "vdW": Quantity("0.1 kilojoule_per_mole"),
            "Electrostatics": Quantity("0.5 kilojoule_per_mole"),
        },
    )


@skip_if_missing("openmm")
@pytest.mark.parametrize("valence_handler", ["Bonds", "Angles", "ProperTorsions", "ImproperTorsions"])
def test_issue_1433_openmm(sage, valence_handler):
    """Test that changes in Collection.set_force_field_parameters are reflected in re-computed energies."""
    interchange = sage.create_interchange(
        MoleculeWithConformer.from_smiles(
            "CC(=O)NC",
            allow_undefined_stereo=True,
        ).to_topology(),
    )

    valence_parameters = interchange[valence_handler].get_force_field_parameters()

    original_energy = get_openmm_energies(interchange).total_energy.m

    # multiple force constant (0th column for each handler) by arbtirarily large number
    valence_parameters[:, 0] = valence_parameters[:, 0] * 100

    # impropers do basically nothing energetically, so also shift phase by a few degrees
    # to force an energy change
    if valence_handler == "ImproperTorsions":
        valence_parameters[:, 2] = valence_parameters[:, 2] + 10  # implicit degrees

    interchange[valence_handler].set_force_field_parameters(valence_parameters)

    new_energy = get_openmm_energies(interchange).total_energy.m

    assert abs(original_energy - new_energy) > 0.1


@pytest.mark.parametrize("engine", ["openmm", "amber", "gromacs"])
def test_issue_1450_reassign_vdw(sage, water_dimer, engine):
    """Sanity check that re-assigning vdW parameters actually changes energies."""

    if engine == "amber" and not has_executable("sander"):
        pytest.skip(reason="sander not installed")

    pytest.importorskip("openmm")

    original_interchange = sage.create_interchange(water_dimer)

    match engine:
        case "openmm":
            original_energy = get_openmm_energies(original_interchange, combine_nonbonded_forces=False).energies["vdW"]
        case "amber":
            original_energy = get_amber_energies(original_interchange).energies["vdW"]
        case "gromacs":
            original_energy = get_gromacs_energies(original_interchange).energies["vdW"]

    for potential in original_interchange["vdW"].potentials.values():
        potential.parameters["epsilon"] = potential.parameters["epsilon"] * 100

    match engine:
        case "openmm":
            new_energy = get_openmm_energies(original_interchange, combine_nonbonded_forces=False).energies["vdW"]
        case "amber":
            new_energy = get_amber_energies(original_interchange).energies["vdW"]
        case "gromacs":
            new_energy = get_gromacs_energies(original_interchange).energies["vdW"]

    assert original_energy != new_energy, "Energies should be different after modifying vdW parameters"


@pytest.mark.parametrize("engine", ["openmm", "amber", "gromacs"])
@pytest.mark.parametrize("valence_handler", ["Bonds", "Angles", "ProperTorsions"])
def test_issue_1450_reassign_valence(sage, valence_handler, engine):
    """Sanity check that re-assigning valence parameters actually changes energies."""

    pytest.importorskip("openmm")

    if engine == "amber" and not has_executable("sander"):
        pytest.skip(reason="sander not installed")

    topology = MoleculeWithConformer.from_smiles(
        "CC(=O)NC",
        allow_undefined_stereo=True,
    ).to_topology()

    topology.box_vectors = Quantity(numpy.eye(3) * 10.0, "nanometer")

    original_interchange = sage.create_interchange(topology)

    match valence_handler:
        case "Bonds":
            shorthand = "Bond"
        case "Angles":
            shorthand = "Angle"
        case "ProperTorsions":
            shorthand = "Torsion"

    match engine:
        case "openmm":
            original_energy = get_openmm_energies(original_interchange).energies[shorthand]
        case "amber":
            original_energy = get_amber_energies(original_interchange).energies[shorthand]
        case "gromacs":
            original_energy = get_gromacs_energies(original_interchange).energies[shorthand]

    for potential in original_interchange[valence_handler].potentials.values():
        potential.parameters["k"] = potential.parameters["k"] * 100

        if "phase" in potential.parameters:
            potential.parameters["phase"] = potential.parameters["phase"] + Quantity(20, "degree")

    match engine:
        case "openmm":
            new_energy = get_openmm_energies(original_interchange).energies[shorthand]
        case "amber":
            new_energy = get_amber_energies(original_interchange).energies[shorthand]
        case "gromacs":
            new_energy = get_gromacs_energies(original_interchange).energies[shorthand]

    assert original_energy != new_energy, "Energies should be different after modifying valence parameters"


@skip_if_missing("openmm")
@pytest.mark.parametrize("valence_handler", ["Bonds", "Angles", "ProperTorsions", "ImproperTorsions"])
def test_issue_1234_openmm(sage, valence_handler):
    """Test that modifications to a `ForceField` object are reflected in re-creating new `Interchange`s."""
    topology = MoleculeWithConformer.from_smiles(
        "CC(=O)NC",
        allow_undefined_stereo=True,
    ).to_topology()
    original_interchange = sage.create_interchange(topology)

    original_energy = get_openmm_energies(original_interchange).total_energy.m

    for parameter in sage[valence_handler].parameters:
        if hasattr(parameter, "k"):
            if isinstance(parameter.k, list):
                parameter.k = [k * 100 for k in parameter.k]
            else:
                parameter.k = parameter.k * 100

        # Impropers contribute basically nothing to energies, so modifying the force constant doesn't do much.
        # Instead, shift the phase by 5 degrees, which should have a much more significant effect on energies.
        if hasattr(parameter, "phase"):
            parameter.phase = [phase + Quantity("5 degree") for phase in parameter.phase]
        elif hasattr(parameter, "length"):
            parameter.length = parameter.length * 1.2
        elif hasattr(parameter, "angle"):
            parameter.angle = parameter.angle * 1.1
        else:
            raise ValueError(f"Don't know how to modify parameter with k of type {type(parameter.k)}")

    new_interchange = sage.create_interchange(topology)

    new_energy = get_openmm_energies(new_interchange).total_energy.m

    # energies should be different, since parameters are (wildly!) different
    assert abs(original_energy - new_energy) > 0.1


def test_clear_caches_with_dead_weakref_proxy():
    """Reproduce a bug where _clear_caches raises ReferenceError when dead
    weakref proxies exist in gc.get_objects(). See
    https://github.com/openforcefield/openff-interchange/issues/1453
    """
    import gc
    import weakref

    class _Dummy:
        pass

    # Create a dead weakref proxy that remains tracked by gc
    obj = _Dummy()
    proxy = weakref.proxy(obj)
    container = [proxy]  # noqa: F841
    del obj
    gc.collect()

    molecule = MoleculeWithConformer.from_smiles("C")
    force_field = ForceField("openff-2.2.0.offxml")

    # This raised ReferenceError before the fix
    Interchange.from_smirnoff(force_field=force_field, topology=[molecule])


def test_issue_1461(sage, opc, ethanol):
    with warnings.catch_warnings():
        warnings.simplefilter("error", PresetChargesAndVirtualSitesWarning)

        sage_opc = sage.combine(opc)
        sage_opc.create_interchange(ethanol.to_topology(), charge_from_molecules=[])


@pytest.mark.parametrize("attribute", ["box", "positions"])
def test_issue_1480(sage, ethanol, attribute):
    """Test that positions and box vectors are stored in nm after being set with non-nm length units."""
    out = sage.create_interchange(ethanol.to_topology())

    match attribute:
        case "box":
            out.box = Quantity(numpy.eye(3) * 35, "angstrom")
        case "positions":
            out.positions = Quantity(
                _rng.random(
                    (out.topology.n_atoms, 3),
                ),
                "angstrom",
            )

    assert str(getattr(out, attribute).units) == "nanometer"

    if attribute == "box":
        box = out.box

        assert box[0][0] == box[1][1] == box[2][2]

        assert str(box[0][0].units) == "nanometer"
        assert box[0][0].m == pytest.approx(3.5)
        assert str(box[1][0].units) == "nanometer"
        assert box[1][0].m == pytest.approx(0.0)



from openff.toolkit import Quantity

from openff.interchange._tests import MoleculeWithConformer, needs_gmx, needs_lmp
from openff.interchange.drivers import get_gromacs_energies, get_openmm_energies


class TestLJPME:
    @needs_gmx
    @needs_lmp
    def test_ljpme(self, sage):
        sage["vdW"].periodic_method = "Ewald3D"

        topology = MoleculeWithConformer.from_smiles("CC").to_topology()
        topology.box_vectors = Quantity([10, 10, 10], "nanometer")
        interchange = sage.create_interchange(topology)

        get_openmm_energies(interchange, combine_nonbonded_forces=True).compare(
            get_gromacs_energies(interchange),
            tolerances={
                "Nonbonded": Quantity(0.05, "kilojoule / mole"),
            },
        )



import pytest
from openff.toolkit import Molecule, Quantity
from openff.toolkit.typing.engines.smirnoff.parameters import GBSAHandler

from openff.interchange.constants import kcal_mol_a2
from openff.interchange.models import TopologyKey
from openff.interchange.smirnoff._gbsa import SMIRNOFFGBSACollection


class TestGBSACollection:
    @pytest.fixture
    def gbsa_handler(self) -> GBSAHandler:
        handler = GBSAHandler(
            version="0.3",
            gb_model="OBC1",
            solvent_dielectric="78.5",
            solute_dielectric="1.0",
            sa_model="ACE",
            surface_area_penalty=Quantity(5.4, "kcal/mol/angstrom^2"),
            solvent_radius=Quantity(1.4, "angstrom"),
        )

        for parameter in [
            {"smirks": "[#1:1]", "radius": "0.12 * nanometer", "scale": 0.85},
            {"smirks": "[#1:1]~[#6]", "radius": "0.13 * nanometer", "scale": 0.85},
            {"smirks": "[#1:1]~[#8]", "radius": "0.08 * nanometer", "scale": 0.85},
            {"smirks": "[#1:1]~[#16]", "radius": "0.08 * nanometer", "scale": 0.85},
            {"smirks": "[#6:1]", "radius": "0.22 * nanometer", "scale": 0.72},
            {"smirks": "[#7:1]", "radius": "0.155 * nanometer", "scale": 0.79},
            {"smirks": "[#8:1]", "radius": "0.15 * nanometer", "scale": 0.85},
            {"smirks": "[#9:1]", "radius": "0.15 * nanometer", "scale": 0.88},
            {"smirks": "[#14:1]", "radius": "0.21 * nanometer", "scale": 0.8},
            {"smirks": "[#15:1]", "radius": "0.185 * nanometer", "scale": 0.86},
            {"smirks": "[#16:1]", "radius": "0.18 * nanometer", "scale": 0.96},
            {"smirks": "[#17:1]", "radius": "0.17 * nanometer", "scale": 0.8},
        ]:
            handler.add_parameter(parameter)

        return handler

    def test_create_basic(self, gbsa_handler):
        collection = SMIRNOFFGBSACollection.create(
            gbsa_handler,
            Molecule.from_mapped_smiles(
                "[H:4][C:1]([H:5])([H:6])[C:2]([H:7])([H:8])[O:3][H:9]",
            ).to_topology(),
        )

        assert collection.gb_model == "OBC1"
        assert collection.solvent_dielectric == 78.5
        assert collection.solute_dielectric == 1.0
        assert collection.sa_model == "ACE"
        assert collection.surface_area_penalty == 5.4 * kcal_mol_a2
        assert collection.solvent_radius == Quantity(1.4, "angstrom")

        assert len(collection.key_map) == 9
        assert len(collection.potentials) == 4

        oxygen_parameters = collection.potentials[collection.key_map[TopologyKey(atom_indices=(2,))]].parameters

        assert oxygen_parameters["radius"] == Quantity(0.15, "nanometer")
        assert oxygen_parameters["scale"] == Quantity(0.85, "dimensionless")



import numpy
import pytest
from openff.toolkit import Molecule, Quantity, Topology

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer
from openff.interchange.interop.gromacs._interchange import (
    _convert_topology,
    to_interchange,
)
from openff.interchange.interop.gromacs.models.models import GROMACSSystem
from openff.interchange.smirnoff._gromacs import _convert


class TestToInterchange:
    @pytest.fixture
    def simple_interchange(self, sage_unconstrained) -> Interchange:
        topology = Topology()
        for index, (smiles, name) in enumerate(
            zip(
                [
                    "[H:2][C:1]([H:3])([H:4])[H:5]",
                    "[H:3][C:1](=[C:2]([H:5])[H:6])[H:4]",
                    "[H:2][O:1][H:3]",
                ],
                ["MET", "ETH", "WAT"],
            ),
        ):
            molecule = Molecule.from_mapped_smiles(smiles)
            molecule.generate_conformers(n_conformers=1)
            molecule.conformers[0] += Quantity(numpy.array([index, 0, 0]), "nanometer")
            molecule.name = name

            topology.add_molecule(molecule)

        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        return sage_unconstrained.create_interchange(topology)

    def test_convert_basic_system(self, monkeypatch, simple_interchange):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

        converted = to_interchange(_convert(simple_interchange))

        assert numpy.allclose(converted.positions, simple_interchange.positions)
        assert numpy.allclose(converted.box, simple_interchange.box)

        for name in simple_interchange.collections:
            # Currently two bond potentials (for rigid water) are lost in the conversion;
            # not clear if this information should be retained.
            # * bond parameters would need to be written out to [ bonds ] even with [ settles ] populated
            # * GromacsMolecule.bonds would need to be populated even if it has a settles
            # * the collection would need to store the bond parameters despite also being constrained

            # Similar situation with angles

            assert name in converted.collections


class TestConvertTopology:
    @pytest.fixture
    def simple_system(self, sage_unconstrained) -> GROMACSSystem:
        molecule = Molecule.from_smiles("CCO")
        molecule.generate_conformers(n_conformers=1)
        topology = Topology.from_molecules([molecule])
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        return _convert(sage_unconstrained.create_interchange(topology))

    @pytest.fixture
    def water_dimer(self, sage_unconstrained, water):
        water.name = "WAT"
        topology = Topology.from_molecules([water, water])
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        return _convert(sage_unconstrained.create_interchange(topology))

    def test_convert_basic_system(self, monkeypatch, simple_system):
        monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")
        converted = _convert_topology(simple_system)

        assert converted.n_molecules == 1
        assert converted.n_atoms == 9
        assert converted.n_bonds == 8

    def test_water_with_settles_has_bonds_in_topology(self, water_dimer):
        assert "WAT" in water_dimer.molecule_types
        assert len(water_dimer.molecule_types["WAT"].settles) == 1

        converted = _convert_topology(water_dimer)

        assert converted.molecule(0).name == "WAT"
        assert converted.molecule(1).name == "WAT"
        assert converted.n_molecules == 2
        assert converted.n_atoms == 6
        assert converted.n_bonds == 4

        assert [atom.atomic_number for atom in converted.atoms] == [8, 1, 1, 8, 1, 1]

    def test_nongrouped_molecules(self, sage):
        """Test that a system containing un-grouped identical molecules is handled correctly."""

        pytest.importorskip("openmm")

        import openmm.app

        sage.create_interchange(
            Topology.from_molecules(
                [
                    MoleculeWithConformer.from_smiles("O"),
                    MoleculeWithConformer.from_smiles("N"),
                    MoleculeWithConformer.from_smiles("O"),
                ],
            ),
        ).to_top("ungrouped.top")

        assert [
            "MOL0",
            "MOL1",
            "MOL0",
        ] == [residue.name for residue in openmm.app.GromacsTopFile("ungrouped.top").topology.residues()]



from openff.toolkit import Quantity

from openff.interchange.common._nonbonded import ElectrostaticsCollection
from openff.interchange.models import (
    LibraryChargeTopologyKey,
    TopologyKey,
    VirtualSiteKey,
)


class BasicElectrostaticsCollection(ElectrostaticsCollection):
    """A slightly more complete collection than the base class."""

    @property
    def charges(
        self,
    ) -> dict[TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey, Quantity]:
        """Get the total partial charge on each atom, including virtual sites."""
        if len(self._charges) == 0 or self._charges_cached is False:
            self._charges = self._get_charges()
            self._charges_cached = True

        return self._charges

    def _get_charges(  # type: ignore
        self,
    ) -> dict[
        TopologyKey | LibraryChargeTopologyKey | VirtualSiteKey,
        Quantity,
    ]:
        charges: dict[TopologyKey | LibraryChargeTopologyKey, Quantity] = dict()

        for topology_key, potential_key in self.key_map.items():
            potential = self.potentials[potential_key]

            charges[topology_key] = potential.parameters["charge"]

        return charges



from collections.abc import Sequence
from typing import Any

class Particle:
    def __init__(self, name: str | None) -> None: ...

class Compound:
    def __init__(
        self,
        name: str = "Compound",
        subcompounds: list | None = None,
    ) -> None: ...
    name: str
    xyz: Any
    def add(
        self,
        new_child: Any,
        label: str | None = None,
        containment: bool = True,
        replace: bool = False,
        inherit_periodicity: bool = True,
        inherit_box: bool = False,
        reset_rigid_ids: bool = True,
    ) -> None: ...
    def add_bond(self, particle_pair: Sequence) -> None: ...
    def __getitem__(self, selection: int) -> Compound: ...



import random

import pytest
from openff.toolkit import Topology
from openff.toolkit.typing.engines.smirnoff.parameters import (
    AngleHandler,
    ParameterHandler,
)

from openff.interchange.exceptions import InvalidParameterHandlerError
from openff.interchange.smirnoff import (
    SMIRNOFFAngleCollection,
    SMIRNOFFCollection,
    SMIRNOFFElectrostaticsCollection,
)


class TestSMIRNOFFCollection:
    def test_allowed_parameter_handler_types(self):
        class DummyParameterHandler(ParameterHandler):
            pass

        class DummySMIRNOFFCollection(SMIRNOFFCollection):
            type: str = "Dummy"
            expression: str = "1+1"

            @classmethod
            def allowed_parameter_handlers(cls):
                return [DummyParameterHandler]

            @classmethod
            def supported_parameters(cls):
                return list()

        dummy_handler = DummySMIRNOFFCollection()
        angle_Handler = AngleHandler(version=0.3)

        assert DummyParameterHandler in dummy_handler.allowed_parameter_handlers()
        assert AngleHandler not in dummy_handler.allowed_parameter_handlers()
        assert DummyParameterHandler not in SMIRNOFFAngleCollection.allowed_parameter_handlers()

        dummy_handler = DummyParameterHandler(version=0.3)

        with pytest.raises(InvalidParameterHandlerError):
            SMIRNOFFAngleCollection.create(
                parameter_handler=dummy_handler,
                topology=Topology(),
            )

        with pytest.raises(InvalidParameterHandlerError):
            DummySMIRNOFFCollection.create(
                parameter_handler=angle_Handler,
                topology=Topology(),
            )


def test_json_roundtrip_preserves_float_values():
    """Reproduce issue #908."""
    scale_factor = 0.5 + random.random() * 0.5

    collection = SMIRNOFFElectrostaticsCollection(scale_14=scale_factor)

    assert collection.scale_14 == scale_factor

    roundtripped = SMIRNOFFElectrostaticsCollection.model_validate_json(
        collection.model_dump_json(),
    )

    assert roundtripped.scale_14 == scale_factor



from typing import Literal

from openff.toolkit import Topology
from pydantic import Field

from openff.interchange.common._valence import (
    AngleCollection,
    BondCollection,
    ProperTorsionCollection,
    RyckaertBellemansTorsionCollection,
)
from openff.interchange.foyer._base import (
    POTENTIAL_KEY_SEPARATOR,
    FoyerConnectedAtomsHandler,
    _copy_params,
    _get_potential_key_id,
)
from openff.interchange.models import PotentialKey, TopologyKey


class FoyerHarmonicBondHandler(FoyerConnectedAtomsHandler, BondCollection):
    """Handler storing bond potentials as produced by a Foyer force field."""

    type: Literal["Bonds"] = "Bonds"
    expression: str = "k/2*(r-length)**2"  # type: ignore[assignment]
    force_field_key: str = "harmonic_bonds"
    connection_attribute: str = "bonds"

    def get_params_with_units(self, params):
        """Get the parameters of this handler, tagged with units."""
        return _copy_params(
            params,
            param_units={"k": "kJ / mol / nm**2", "length": "nm"},
        )

    def store_matches(
        self,
        atom_slots: dict[TopologyKey, PotentialKey],
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""
        for bond in topology.bonds:
            atom_indices = (
                topology.atom_index(bond.atom1),
                topology.atom_index(bond.atom2),
            )
            top_key = TopologyKey(atom_indices=atom_indices)

            pot_key_ids = tuple(_get_potential_key_id(atom_slots, idx) for idx in atom_indices)

            self.key_map[top_key] = PotentialKey(
                id=POTENTIAL_KEY_SEPARATOR.join(pot_key_ids),
            )


class FoyerHarmonicAngleHandler(FoyerConnectedAtomsHandler, AngleCollection):
    """Handler storing angle potentials as produced by a Foyer force field."""

    type: Literal["Angles"] = "Angles"
    expression: str = "k/2*(theta-angle)**2"  # type: ignore[assignment]
    force_field_key: str = "harmonic_angles"
    connection_attribute: str = "angles"

    def get_params_with_units(self, params):
        """Get the parameters of this handler, tagged with units."""
        return _copy_params(
            {"k": params["k"], "angle": params["theta"]},
            param_units={
                "k": "kJ / mol / radian**2",
                "angle": "dimensionless",  # should this really be dimensionless ... ?
            },
        )

    def store_matches(
        self,
        atom_slots: dict[TopologyKey, PotentialKey],
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""
        for angle in topology.angles:
            atom_indices = tuple(topology.atom_index(atom) for atom in angle)
            top_key = TopologyKey(atom_indices=atom_indices)

            pot_key_ids = tuple(_get_potential_key_id(atom_slots, idx) for idx in atom_indices)

            self.key_map[top_key] = PotentialKey(
                id=POTENTIAL_KEY_SEPARATOR.join(pot_key_ids),
            )


class FoyerRBProperHandler(
    FoyerConnectedAtomsHandler,
    RyckaertBellemansTorsionCollection,
):
    """Handler storing Ryckaert-Bellemans proper torsion potentials as produced by a Foyer force field."""

    force_field_key: str = "rb_propers"
    type: Literal["RBTorsions"] = "RBTorsions"
    expression: str = Field(
        "c0 + "
        "c1 * (cos(phi - 180)) "
        "c2 * (cos(phi - 180)) ** 2 + "
        "c3 * (cos(phi - 180)) ** 3 + "
        "c4 * (cos(phi - 180)) ** 4 + "
        "c5 * (cos(phi - 180)) ** 5",
    )
    connection_attribute: str = "propers"
    raise_on_missing_params: bool = False

    def get_params_with_units(self, params):
        """Get the parameters of this handler, tagged with units."""
        rb_params = params
        param_units = {k: "kJ / mol" for k in rb_params}
        return _copy_params(rb_params, param_units=param_units)

    def store_matches(
        self,
        atom_slots: dict[TopologyKey, PotentialKey],
        topology: "Topology",
    ) -> None:
        """Populate self.key_map with key-val pairs of [TopologyKey, PotentialKey]."""
        for proper in topology.propers:
            atom_indices = tuple(topology.atom_index(atom) for atom in proper)
            top_key = TopologyKey(atom_indices=atom_indices)

            pot_key_ids = tuple(_get_potential_key_id(atom_slots, idx) for idx in atom_indices)

            self.key_map[top_key] = PotentialKey(
                id=POTENTIAL_KEY_SEPARATOR.join(pot_key_ids),
            )


class FoyerRBImproperHandler(FoyerRBProperHandler):
    """Handler storing Ryckaert-Bellemans improper torsion potentials as produced by a Foyer force field."""

    type: Literal["RBImpropers"] = "RBImpropers"  # type: ignore[assignment]
    connection_attribute: str = "impropers"


class FoyerPeriodicProperHandler(FoyerConnectedAtomsHandler, ProperTorsionCollection):
    """Handler storing periodic proper torsion potentials as produced by a Foyer force field."""

    force_field_key: str = "periodic_propers"
    connection_attribute: str = "propers"
    raise_on_missing_params: bool = False
    type: str = "ProperTorsions"  # type: ignore[assignment]
    expression: str = "k*(1+cos(periodicity*theta-phase))"  # type: ignore[assignment]

    def get_params_with_units(self, params):
        """Get the parameters of this handler, tagged with units."""
        return _copy_params(
            params,
            param_units={
                "k": "kJ / mol / nm**2",
                "phase": "dimensionless",
                "periodicity": "dimensionless",
            },
        )


class FoyerPeriodicImproperHandler(FoyerPeriodicProperHandler):
    """Handler storing periodic improper torsion potentials as produced by a Foyer force field."""

    type: str = "ImproperTorsions"
    connection_attribute: str = "impropers"


class _RBTorsionHandler(RyckaertBellemansTorsionCollection):
    pass



import numpy
import pytest

from openff.interchange._tests import get_test_file_path
from openff.interchange.exceptions import GROMACSParseError
from openff.interchange.interop.gromacs._import._import import from_files
from openff.interchange.interop.gromacs.models.models import RyckaertBellemansDihedral


def test_load_rb_torsions(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    system = from_files(
        get_test_file_path("ethanol_rb_torsions.top"),
        get_test_file_path("ethanol_rb_torsions.gro"),
    )

    assert len(system.molecule_types["Compound"].dihedrals) > 0

    for torsion in system.molecule_types["Compound"].dihedrals:
        assert isinstance(torsion, RyckaertBellemansDihedral)


def test_parse_bad_characters(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    with pytest.raises(
        GROMACSParseError,
        match=r"garbage.*FLAG 8EE0CCA23E",
    ):
        # gro file doesn't need to have contents, crash should happen on the
        # topology file which happens to be parsed first

        from_files(
            get_test_file_path("invalid.top"),
            get_test_file_path("invalid.gro"),
        )


def test_asterisks_are_comments(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    imported = from_files(
        get_test_file_path("asterisk.top"),
        get_test_file_path("asterisk.gro"),
    )

    # make sure asterisk in atom type is not lost
    assert imported.molecule_types["MOL0"].atoms[1].atom_type.endswith("*")
    assert imported.molecule_types["MOL0"].atoms[2].atom_type.endswith("*")

    # just check a few details, non-exhaustive
    assert imported.vdw_14 == 0.5
    assert imported.molecule_types["MOL0"].atoms[0].charge.m == -0.834
    assert numpy.allclose(imported.box.m_as("nanometer"), 3 * numpy.eye(3))


def test_parse_invalid_directive_partway_through(monkeypatch):
    monkeypatch.setenv("INTERCHANGE_EXPERIMENTAL", "1")

    with pytest.raises(
        GROMACSParseError,
        match=r"directive.*flag",
    ):
        from_files(
            get_test_file_path("invalid_partway_through.top"),
            get_test_file_path("invalid.gro"),
        )



import pytest
from nonbonded_plugins.nonbonded import (
    BuckinghamHandler,
    DoubleExponentialHandler,
    SMIRNOFFBuckinghamCollection,
    SMIRNOFFDoubleExponentialCollection,
)
from openff.toolkit import ForceField, Molecule
from openff.toolkit.typing.engines.smirnoff.plugins import load_handler_plugins

from openff.interchange import Interchange
from openff.interchange._tests import get_test_file_path
from openff.interchange.plugins import load_smirnoff_plugins


def test_load_handler_plugins():
    # This does not test Interchange per se but this behavior is necessary for the other tests to function
    available_plugins = load_handler_plugins()

    assert BuckinghamHandler in available_plugins
    assert DoubleExponentialHandler in available_plugins


def test_load_smirnoff_plugins():
    available_plugins = load_smirnoff_plugins()

    assert SMIRNOFFBuckinghamCollection in available_plugins
    assert SMIRNOFFDoubleExponentialCollection in available_plugins


class TestDoubleExponential:
    pytest.importorskip("openmm")

    @pytest.fixture
    def de_force_field(self) -> ForceField:
        force_field = ForceField(
            get_test_file_path("de-force-1.0.1.offxml"),
            load_plugins=True,
        )

        # This force field also includes a 4-site water model, remove that for now
        force_field.deregister_parameter_handler("VirtualSites")
        force_field.deregister_parameter_handler("LibraryCharges")
        force_field["OtherDoubleExponential"].parameters.pop(-1)
        force_field["OtherDoubleExponential"].parameters.pop(-1)
        force_field["Constraints"].parameters.pop(-1)
        force_field["Constraints"].parameters.pop(-1)

        return force_field

    def test_loadable(self):
        ForceField(
            get_test_file_path("de-force-1.0.1.offxml"),
            load_plugins=True,
        )

    def test_create_interchange(self, de_force_field):
        Interchange.from_smirnoff(
            de_force_field,
            Molecule.from_smiles("CO").to_topology(),
        )



import pathlib
import warnings
from typing import IO

import numpy
from openff.toolkit import Quantity

from openff.interchange.exceptions import MissingPositionsError, NonperiodicNoCutoffNotSupportedError
from openff.interchange.interop.gromacs.models.models import (
    GROMACSSystem,
    GROMACSVirtualSite2,
    GROMACSVirtualSite3,
    GROMACSVirtualSite3fad,
    LennardJonesAtomType,
    PeriodicImproperDihedral,
    PeriodicProperDihedral,
    RyckaertBellemansDihedral,
)
from openff.interchange.pydantic import _BaseModel
from openff.interchange.warnings import MissingPositionsWarning


class GROMACSWriter(_BaseModel):
    """Thin wrapper for writing GROMACS systems."""

    system: GROMACSSystem
    top_file: pathlib.Path | str
    gro_file: pathlib.Path | str

    def to_top(self, monolithic: bool = True, _merge_atom_types: bool = False):
        """Write a GROMACS topology file."""
        with open(self.top_file, "w") as top:
            self._write_defaults(top)

            if monolithic:
                atomtypes_file_object: IO[str] = top
            else:
                prefix = str(self.top_file).split(".top")[0]
                atomtypes_file_object = open(
                    f"{prefix}_atomtypes.itp",
                    "w",
                )
                top.write(f'#include "{prefix}_atomtypes.itp"\n')

            mapping_to_reduced_atom_types = self._write_atomtypes(
                top=atomtypes_file_object,
                merge_atom_types=_merge_atom_types,
            )

            self._write_moleculetypes(
                top=top,
                monolithic=monolithic,
                mapping_to_reduced_atom_types=mapping_to_reduced_atom_types,
                merge_atom_types=_merge_atom_types,
            )

            self._write_system(top)
            self._write_molecules(top)

            if not monolithic:
                atomtypes_file_object.close()

    def to_gro(self, decimal: int = 3):
        """Write a GROMACS coordinate file."""
        if self.gro_file is None:
            raise ValueError("No GRO file specified.")

        with open(self.gro_file, "w") as gro:
            self._write_gro(gro, decimal)

    def _write_defaults(self, top: IO[str]):
        top.write("[ defaults ]\n")
        top.write("; nbfunc\tcomb-rule\tgen-pairs\tfudgeLJ\tfudgeQQ\n")

        gen_pairs = "yes" if self.system.gen_pairs else "no"
        top.write(
            f"{self.system.nonbonded_function:6d}\t"
            f"{self.system.combination_rule:6d}\t"
            f"{gen_pairs:6s}\t"
            f"{self.system.vdw_14:8.6f}\t"
            f"{self.system.coul_14:8.6f}\n\n",
        )

    def _write_atomtypes(self, top: IO[str], merge_atom_types: bool) -> dict[str, bool | str]:
        top.write("[ atomtypes ]\n")
        top.write(
            ";type, bondingtype, atomic_number, mass, charge, ptype, sigma, epsilon\n",
        )

        reduced_atom_types: list[tuple[str, LennardJonesAtomType]] = list()
        mapping_to_reduced_atom_types = dict()

        def _is_atom_type_in_list(
            atom_type,
            atom_type_list,
        ) -> bool | str:
            """
            Check if the atom type is already in list.
            """
            mass_tolerance = Quantity("1e-5 amu")
            sigma_tolerance = Quantity("1e-5 nanometer")
            epsilon_tolerance = Quantity("1e-5 kilojoule_per_mole")
            for _at_name, _atom_type in atom_type_list:
                if (
                    atom_type.atomic_number == _atom_type.atomic_number
                    and abs(atom_type.mass - _atom_type.mass) < mass_tolerance
                    and abs(atom_type.sigma - _atom_type.sigma) < sigma_tolerance
                    and abs(atom_type.epsilon - _atom_type.epsilon) < epsilon_tolerance
                ):
                    return _at_name
            return False

        def _get_new_entry_name(atom_type_list) -> str:
            """
            Entry name for atom type to be added.
            """
            if len(reduced_atom_types) > 0:
                _previous_name = reduced_atom_types[-1][0]
                _previous_idx = int(_previous_name.split("_")[1])
                return f"AT_{_previous_idx + 1}"
            else:
                return "AT_0"

        for atom_type in self.system.atom_types.values():
            if not isinstance(atom_type, LennardJonesAtomType):
                raise NotImplementedError(
                    "Only Lennard-Jones atom types are currently supported.",
                )

            if merge_atom_types:
                if _is_atom_type_in_list(atom_type, reduced_atom_types):
                    mapping_to_reduced_atom_types[atom_type.name] = _is_atom_type_in_list(
                        atom_type,
                        reduced_atom_types,
                    )
                else:
                    _at_name = _get_new_entry_name(reduced_atom_types)
                    reduced_atom_types.append((_at_name, atom_type))
                    mapping_to_reduced_atom_types[atom_type.name] = _at_name
            else:
                top.write(
                    f"{atom_type.name:<11s}\t"
                    f"{atom_type.atomic_number:6d}\t"
                    f"{atom_type.mass.m:.16g}\t"
                    f"{atom_type.charge.m:.16f}\t"
                    f"{atom_type.particle_type:5s}\t"
                    f"{atom_type.sigma.m:.16g}\t"
                    f"{atom_type.epsilon.m:.16g}\n",
                )

        if not merge_atom_types:
            top.write("\n")
            return mapping_to_reduced_atom_types

        for atom_type_name, atom_type in reduced_atom_types:
            top.write(
                f"{atom_type_name:<11s}\t"
                f"{atom_type.atomic_number:6d}\t"
                f"{atom_type.mass.m:.16g}\t"
                f"{atom_type.charge.m:.16f}\t"
                f"{atom_type.particle_type:5s}\t"
                f"{atom_type.sigma.m:.16g}\t"
                f"{atom_type.epsilon.m:.16g}\n",
            )
        top.write("\n")
        return mapping_to_reduced_atom_types

    def _write_moleculetypes(
        self,
        top: IO[str],
        monolithic: bool,
        mapping_to_reduced_atom_types,
        merge_atom_types: bool,
    ):
        for molecule_name, molecule_type in self.system.molecule_types.items():
            # this string needs to be something that plays nicely in file paths
            # and also works as GROMACS's label for the moleculetype "name"
            canonicalized_name = molecule_name.translate({ord(c): "_" for c in r' \/:<>"|?*'})

            if monolithic:
                molecule_file = top
            else:
                prefix = str(self.top_file).split(".top")[0]
                molecule_file = open(
                    file=f"{prefix}_{canonicalized_name}.itp",
                    mode="w",
                )
                top.write(f'#include "{prefix}_{canonicalized_name}.itp"\n')

            molecule_file.write("[ moleculetype ]\n")

            molecule_file.write(f"{canonicalized_name}\t{molecule_type.nrexcl:10d}\n\n")

            self._write_atoms(
                molecule_file,
                molecule_type,
                mapping_to_reduced_atom_types,
                merge_atom_types,
            )
            self._write_pairs(molecule_file, molecule_type)
            self._write_bonds(molecule_file, molecule_type)
            self._write_angles(molecule_file, molecule_type)
            self._write_dihedrals(molecule_file, molecule_type)
            self._write_settles(molecule_file, molecule_type)
            self._write_virtual_sites(molecule_file, molecule_type)
            self._write_exclusions(molecule_file, molecule_type)

            if not monolithic:
                molecule_file.close()

        top.write("\n")

    def _write_atoms(
        self,
        top,
        molecule_type,
        mapping_to_reduced_atom_types,
        merge_atom_types: bool,
    ):
        top.write("[ atoms ]\n")
        top.write(";index, atom type, resnum, resname, name, cgnr, charge, mass\n")

        for atom in molecule_type.atoms:
            if merge_atom_types:
                top.write(
                    f"{atom.index:6d} "
                    f"{mapping_to_reduced_atom_types[atom.atom_type]:6s}"
                    f"{atom.residue_index % 100_000:8d} "
                    f"{atom.residue_name:8s} "
                    f"{atom.name:6s}"
                    f"{atom.charge_group_number:6d}"
                    f"{atom.charge.m:20.12f}"
                    f"{atom.mass.m:20.12f}\n",
                )
            else:
                top.write(
                    f"{atom.index:6d} "
                    f"{atom.atom_type:6s}"
                    f"{atom.residue_index % 100_000:8d} "
                    f"{atom.residue_name:8s} "
                    f"{atom.name:6s}"
                    f"{atom.charge_group_number:6d}"
                    f"{atom.charge.m:20.12f}"
                    f"{atom.mass.m:20.12f}\n",
                )

        top.write("\n")

    def _write_pairs(self, top, molecule_type):
        top.write("[ pairs ]\n")
        top.write(";ai    aj   funct\n")

        function = 1

        for pair in molecule_type.pairs:
            top.write(
                f"{pair.atom1:6d}\t{pair.atom2:6d}\t{function:6d}\n",
            )

        top.write("\n")

    def _write_bonds(self, top, molecule_type):
        top.write("[ bonds ]\n")
        top.write(";ai    aj   funct r k\n")

        function = 1

        for bond in molecule_type.bonds:
            top.write(
                f"{bond.atom1:6d} {bond.atom2:6d} {function:6d}{bond.length.m:20.12f} {bond.k.m:20.12f} ",
            )

            top.write("\n")

        top.write("\n")

    def _write_angles(self, top, molecule_type):
        top.write("[ angles ]\n")
        top.write(";ai    aj   ak   funct theta  k\n")

        function = 1

        for angle in molecule_type.angles:
            top.write(
                f"{angle.atom1:6d} "
                f"{angle.atom2:6d} "
                f"{angle.atom3:6d} "
                f"{function:6d} "
                f"{angle.angle.m:20.12f} "
                f"{angle.k.m:20.12f} ",
            )

            top.write("\n")

        top.write("\n")

    def _write_dihedrals(self, top, molecule_type):
        top.write("[ dihedrals ]\n")
        top.write(";ai    aj   ak   al   funct phi  k\n")

        functions = {
            PeriodicProperDihedral: 1,
            RyckaertBellemansDihedral: 3,
            PeriodicImproperDihedral: 4,
        }

        for dihedral in molecule_type.dihedrals:
            function = functions[type(dihedral)]

            top.write(
                f"{dihedral.atom1:6d}"
                f"{dihedral.atom2:6d}"
                f"{dihedral.atom3:6d}"
                f"{dihedral.atom4:6d}"
                f"{functions[type(dihedral)]:6d}",
            )

            if function in [1, 4]:
                top.write(
                    f"{dihedral.phi.m:20.12f}{dihedral.k.m:20.12f}{dihedral.multiplicity:18d}",
                )

            elif function == 3:
                top.write(
                    f"{dihedral.c0.m:20.12f}"
                    f"{dihedral.c1.m:20.12f}"
                    f"{dihedral.c2.m:20.12f}"
                    f"{dihedral.c3.m:20.12f}"
                    f"{dihedral.c4.m:20.12f}"
                    f"{dihedral.c5.m:20.12f}",
                )

            else:
                raise ValueError(f"Invalid dihedral function {function}.")

            top.write("\n")

        top.write("\n")

    def _write_virtual_sites(self, top, molecule_type):
        # TODO: Collect by type so only one header is used for each per molecule
        for gromacs_virtual_site in molecule_type.virtual_sites:
            if isinstance(gromacs_virtual_site, GROMACSVirtualSite2):
                top.write("[ virtual_sites2 ]\n")
                top.write("; parent, orientation atoms, func, a\n")
                top.write(
                    f"{gromacs_virtual_site.site}\t"
                    f"{gromacs_virtual_site.orientation_atoms[0]}\t"
                    f"{gromacs_virtual_site.orientation_atoms[1]}\t"
                    f"{gromacs_virtual_site.func}\t"
                    f"{gromacs_virtual_site.a}\t"
                    "\n",
                )

            elif isinstance(gromacs_virtual_site, GROMACSVirtualSite3):
                top.write("[ virtual_sites3 ]\n")
                top.write("; parent, orientation atoms, func, a, b\n")
                top.write(
                    f"{gromacs_virtual_site.site}\t"
                    f"{gromacs_virtual_site.orientation_atoms[0]}\t"
                    f"{gromacs_virtual_site.orientation_atoms[1]}\t"
                    f"{gromacs_virtual_site.orientation_atoms[2]}\t"
                    f"{gromacs_virtual_site.func}\t"
                    f"{gromacs_virtual_site.a}\t"
                    f"{gromacs_virtual_site.b}\t"
                    "\n",
                )

            elif isinstance(gromacs_virtual_site, GROMACSVirtualSite3fad):
                top.write("[ virtual_sites3 ]\n")
                top.write("; parent, orientation atoms, func, theta, d\n")
                top.write(
                    f"{gromacs_virtual_site.site}\t"
                    f"{gromacs_virtual_site.orientation_atoms[0]}\t"
                    f"{gromacs_virtual_site.orientation_atoms[1]}\t"
                    f"{gromacs_virtual_site.orientation_atoms[2]}\t"
                    f"{gromacs_virtual_site.func}\t"
                    f"{gromacs_virtual_site.theta}\t"
                    f"{gromacs_virtual_site.d}\t"
                    "\n",
                )

            else:
                raise NotImplementedError()

        top.write("\n")

    def _write_exclusions(self, top, molecule_type):
        top.write("[ exclusions ]\n")
        top.write(";ai    aj\n")

        for exclusion in molecule_type.exclusions:
            top.write(
                f"{exclusion.first_atom:6d}",
            )
            for other_atom in exclusion.other_atoms:
                top.write(
                    f"{other_atom:6d}",
                )

            top.write("\n")

        top.write("\n")

    def _write_settles(self, top, molecule_type):
        top.write("[ settles ]\n")
        top.write(";i  funct   dOH  dHH\n")

        function = 1

        for settle in molecule_type.settles:
            top.write(
                f"{settle.first_atom:6d}\t"
                f"{function:6d}\t"
                f"{settle.oxygen_hydrogen_distance.m_as('nanometer'):20.12f}\t"
                f"{settle.hydrogen_hydrogen_distance.m_as('nanometer'):20.12f}\n",
            )

        top.write("\n")

    def _write_system(self, top):
        top.write("[ system ]\n")
        top.write(";name\n")

        top.write(f"{self.system.name}\n")

        top.write("\n")

    def _write_molecules(self, top):
        top.write("[ molecules ]\n")
        top.write(";name\tnumber\n")

        for molecule_name, n_molecules in self.system.molecules:
            top.write(f"{molecule_name.replace(' ', '_')}\t{n_molecules}\n")

        top.write("\n")

    def _write_gro(self, gro, decimal: int):
        if self.system.positions is None:
            raise MissingPositionsError(
                "Positions are required to write a `.gro` file but found None.",
            )
        elif numpy.allclose(self.system.positions, 0):
            warnings.warn(
                "Positions seem to all be zero. Result coordinate file may be non-physical.",
                MissingPositionsWarning,
            )

        n_particles = sum(
            len(self.system.molecule_types[molecule_name].atoms) * n_copies
            for molecule_name, n_copies in self.system.molecules
        )

        assert n_particles == self.system.positions.shape[0], (
            n_particles,
            self.system.positions.shape[0],
        )

        # Explicitly round here to avoid ambiguous things in string formatting
        positions = numpy.round(self.system.positions, decimal).m_as("nanometer")

        gro.write("Generated by Interchange\n")
        gro.write(f"{n_particles}\n")

        count = 0
        for molecule_name, n_copies in self.system.molecules:
            molecule = self.system.molecule_types[molecule_name]

            for copy_index in range(n_copies):
                for atom in molecule.atoms:
                    gro.write(
                        f"%5d%-5s%5s%5d"
                        f"%{decimal + 5}.{decimal}f"
                        f"%{decimal + 5}.{decimal}f"
                        f"%{decimal + 5}.{decimal}f\n"
                        % (
                            (atom.residue_index + copy_index) % 100_000,
                            atom.residue_name[:5],
                            atom.name[:5],
                            (count + 1) % 100000,
                            positions[count, 0],
                            positions[count, 1],
                            positions[count, 2],
                        ),
                    )

                    count += 1

        if self.system.box is None:
            raise NonperiodicNoCutoffNotSupportedError(
                "GROMACS versions 2020 and newer do not support systems without periodicity/box vectors.",
            )
        else:
            box = self.system.box.m_as("nanometer")

        # Check for rectangular
        if (box == numpy.diag(numpy.diagonal(box))).all():
            for i in range(3):
                gro.write(f"{box[i, i]:11.7f}")
        else:
            # v1x v1y v1z
            # v2x v2y v2z
            # v3x v3y v3z
            # https://manual.gromacs.org/archive/5.0.3/online/gro.html
            # v1(x) v2(y) v3(z) v1(y) v1(z) v2(x) v2(z) v3(x) v3(y)
            for value in [
                box[0, 0],  # v1x
                box[1, 1],  # v2y
                box[2, 2],  # v3z
                box[0, 1],  # v1y
                box[0, 2],  # v1z
                box[1, 0],  # v2x
                box[1, 2],  # v2z
                box[2, 0],  # v3x
                box[2, 1],  # v3y
            ]:
                gro.write(f"{value:11.7f}")

        gro.write("\n")



"""
Helper functions for exporting positions to OpenMM.
"""

from typing import TYPE_CHECKING

from openff.interchange.interop.common import _to_positions

if TYPE_CHECKING:
    import openmm.unit

    from openff.interchange import Interchange


def to_openmm_positions(
    interchange: "Interchange",
    include_virtual_sites: bool = True,
) -> "openmm.unit.Quantity":
    """Generate an array of positions of all particles, optionally excluding virtual sites."""
    return _to_positions(interchange, include_virtual_sites).to_openmm()



import copy
from pathlib import Path

import numpy
import pytest
from openff.toolkit import ForceField, Quantity, Topology
from openff.utilities import temporary_cd

from openff.interchange import Interchange
from openff.interchange._tests import MoleculeWithConformer, needs_lmp
from openff.interchange.constants import kj_mol
from openff.interchange.drivers import get_lammps_energies, get_openmm_energies
from openff.interchange.exceptions import EnergyError

rng = numpy.random.default_rng(821)


@needs_lmp
class TestLammps:
    @pytest.mark.parametrize("n_mols", [1, 2])
    @pytest.mark.parametrize(
        "mol",
        [
            "C",
            "CC",  # Adds a proper torsion term(s)
            "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
            "CCOC",  # This hits t86, which has a non-1.0 idivf
            "C1COC(=O)O1",  # This adds an improper, i2
        ],
    )
    def test_to_lammps_single_mols(
        self,
        mol: str,
        sage_unconstrained: ForceField,
        n_mols: int,
    ) -> None:
        """
        Test that Interchange.to_openmm Interchange.to_lammps report sufficiently similar energies.

        TODO: Tighten tolerances
        TODO: Test periodic and non-periodic
        """
        mol = MoleculeWithConformer.from_smiles(mol)
        mol.conformers[0] -= numpy.min(mol.conformers[0], axis=0)
        top = Topology.from_molecules(n_mols * [mol])

        top.box_vectors = Quantity(5.0 * numpy.eye(3), "nanometer")

        if n_mols == 1:
            positions = mol.conformers[0]
        elif n_mols == 2:
            positions = Quantity(
                numpy.concatenate(
                    [mol.conformers[0], mol.conformers[0] + Quantity(1.5, "nanometer")],
                ),
            )

        interchange = Interchange.from_smirnoff(sage_unconstrained, top)
        interchange.positions = positions
        interchange.box = top.box_vectors

        reference = get_openmm_energies(
            interchange=interchange,
            round_positions=3,
        )

        lmp_energies = get_lammps_energies(
            interchange=interchange,
            round_positions=3,
        )

        lmp_energies.compare(
            reference,
            tolerances={
                "Nonbonded": Quantity(1, "kilojoule_per_mole"),
                "Torsion": Quantity(0.02, "kilojoule_per_mole"),
            },
        )

    @pytest.mark.parametrize(
        "n_mols",
        [1, 2],
    )
    @pytest.mark.parametrize(
        "mol",
        [
            "C",
            "CC",  # Adds a proper torsion term(s)
            "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
            "CCOC",  # This hits t86, which has a non-1.0 idivf
            "C1COC(=O)O1",  # This adds an improper, i2
        ],
    )
    def test_to_lammps_single_mols_relative_to_perturbed_coordinates(
        self,
        mol: str,
        sage_unconstrained: ForceField,
        n_mols: int,
    ) -> None:
        """
        Test that Interchange.to_openmm Interchange.to_lammps report sufficiently similar relative energies between
        their initial coordinates and coordinates that have been randomly perturbed.

        TODO: Tighten tolerances
        TODO: Test periodic and non-periodic
        """
        rng = numpy.random.default_rng(821)
        mol = MoleculeWithConformer.from_smiles(mol)
        mol.conformers[0] -= numpy.min(mol.conformers[0], axis=0)
        top = Topology.from_molecules(n_mols * [mol])

        top.box_vectors = Quantity(5.0 * numpy.eye(3), "nanometer")

        if n_mols == 1:
            positions = mol.conformers[0]
        elif n_mols == 2:
            positions = Quantity(
                numpy.concatenate(
                    [mol.conformers[0], mol.conformers[0] + Quantity(1.5, "nanometer")],
                ),
            )

        perturbed_positions = positions + Quantity(rng.random(positions.shape) * 0.1, "angstrom")

        interchange = Interchange.from_smirnoff(sage_unconstrained, top)
        interchange.positions = positions
        interchange.box = top.box_vectors

        perturbed_interchange = copy.deepcopy(interchange)
        perturbed_interchange.positions = perturbed_positions

        reference = get_openmm_energies(
            interchange=interchange,
            round_positions=3,
            combine_nonbonded_forces=False,
        )
        perturbed_reference = get_openmm_energies(
            interchange=perturbed_interchange,
            round_positions=3,
            combine_nonbonded_forces=False,
        )

        lmp_energies = get_lammps_energies(
            interchange=interchange,
            round_positions=3,
        )
        perturbed_lmp_energies = get_lammps_energies(
            interchange=perturbed_interchange,
            round_positions=3,
        )

        reference_energy_differences = {
            key: diff.to(kj_mol) for key, diff in reference.diff(perturbed_reference).items()
        }

        lmp_energy_differences = {
            key: diff.to(kj_mol) for key, diff in lmp_energies.diff(perturbed_lmp_energies).items()
        }

        tolerances = {
            "Bond": 1e-11 * kj_mol,
            "Angle": 1e-10 * kj_mol,
            "Torsion": 1e-12 * kj_mol,
            "vdW": 1e-7 * kj_mol,
            "Electrostatics": 1e-1 * kj_mol,
        }

        errors = dict()
        for key, lmp_diff in lmp_energy_differences.items():
            error = abs(lmp_diff - reference_energy_differences[key])
            if error > tolerances[key]:
                errors[key] = error

        if errors:
            raise EnergyError(errors)

    @pytest.mark.parametrize("n_mols", [2, 3])
    @pytest.mark.parametrize(
        "smiles",
        [
            "CCO",
            "N1CCCC1",
            "c1ccccc1",
        ],
    )
    def test_unique_lammps_mol_ids(
        self,
        smiles,
        sage_unconstrained,
        n_mols,
    ) -> bool:
        """Test to see if interop.lammps.export._write_atoms() writes unique ids for each distinct Molecule"""
        import lammps

        molecule = MoleculeWithConformer.from_smiles(smiles)
        topology = Topology.from_molecules(n_mols * [molecule])

        # Just use random positions since we're testing molecule IDs, not physics
        topology.set_positions(
            Quantity(
                rng.random((topology.n_atoms, 3)),
                "nanometer",
            ),
        )
        topology.box_vectors = Quantity([4, 4, 4], "nanometer")

        with temporary_cd():
            lammps_prefix = Path.cwd() / "lammps_test"

            interchange = sage_unconstrained.create_interchange(topology)
            interchange.to_lammps(lammps_prefix)

            # Extract molecule IDs from data file
            with lammps.lammps(
                cmdargs=["-screen", "none", "-log", "none"],
            ) as lmp:
                lmp.file(
                    "lammps_test_pointenergy.in",
                )
                written_mol_ids = {
                    mol_id
                    for _, mol_id in zip(
                        range(lmp.get_natoms()),
                        lmp.extract_atom("molecule"),
                    )
                }

        # these are expected to be [1...N] for each of N molecules
        expected_mol_ids = {i + 1 for i in range(interchange.topology.n_molecules)}

        assert expected_mol_ids == written_mol_ids

    @pytest.mark.parametrize(
        "mol",
        [
            "C=O",  # Simplest molecule with any improper torsion
            "OC=O",  # Simplest molecule with a multi-term torsion
        ],
    )
    def test_to_lammps_with_type_labels(
        self,
        mol: str,
        sage_unconstrained: ForceField,
        tmp_path,
    ) -> None:
        import lammps

        from openff.interchange.exceptions import LAMMPSRunError

        mol = MoleculeWithConformer.from_smiles(mol)
        mol.conformers[0] -= numpy.min(mol.conformers[0], axis=0)
        top = Topology.from_molecules([mol])

        top.box_vectors = Quantity(5.0 * numpy.eye(3), "nanometer")
        positions = mol.conformers[0]

        interchange = Interchange.from_smirnoff(sage_unconstrained, top)
        interchange.positions = positions
        interchange.box = top.box_vectors

        interchange.to_lammps(tmp_path / "out", include_type_labels=True)

        runner = lammps.lammps(cmdargs=["-screen", "none", "-nocite"])

        try:
            runner.file("out_pointenergy.in")
        except Exception as error:
            raise LAMMPSRunError from error
